
import pytest
from rocketry.pybox.time.convert import to_datetime

from rocketry.time.interval import (
    TimeOfWeek
)

NS_IN_SECOND = 1e+9
NS_IN_MINUTE = 1e+9 * 60
NS_IN_HOUR   = 1e+9 * 60 * 60
NS_IN_DAY    = 1e+9 * 60 * 60 * 24
NS_IN_WEEK   = 1e+9 * 60 * 60 * 24 * 7

# TimeOfWeek
# Year 2024 was chosen as it starts on monday
@pytest.mark.parametrize(
    "dt,string,ns",
    [
        # Regular
        pytest.param(
            to_datetime("2024-01-01 00:00:00"),
            "Mon 00:00:00",
            0,
            id="Beginning"),
        pytest.param(
            to_datetime("2024-01-07 23:59:59.999999000"),
            "Sun 23:59:59.999999000",
            604799999999000.0,
            id="Ending"),
    ],
)
def test_anchor_equal(dt, string, ns):
    time = TimeOfWeek(None, None)
    assert time.anchor_dt(dt) == time.anchor_str(string) == ns


@pytest.mark.parametrize(
    "start,end,start_ns,end_ns",
    [
        # Regular
        pytest.param(
            "Mon 00:00:00",
            "Sun 23:59:59.999999000",
            0, NS_IN_WEEK - 1 - 999, # datetime stuff are often microsecond accuracy
            id="Strings: full"),
        pytest.param(
            # From Tue 00:00 to Wed 23:59:59.000 ()
            "Tue",
            "Wed", # 
            NS_IN_DAY, NS_IN_DAY * 3 - 1,
            id="Strings: minimal"),
    ],
)
def test_construct(start, end, start_ns, end_ns):
    time = TimeOfWeek(start, end)
    assert time._start == start_ns
    assert time._end == end_ns



import datetime
import pytest
from rocketry.time import RelativeDay

@pytest.mark.parametrize("day,left,right", [
    pytest.param('today', datetime.datetime(2022, 5, 15), datetime.datetime(2022, 5, 15, 23, 59, 59, 999999)),  
])
def test_rollback(day, left, right):
    rel_day = RelativeDay(day)

    dt = datetime.datetime(2022, 5, 15)
    interval = rel_day.rollback(dt)
    assert interval.left == left
    assert interval.right == right


def test_rollforward():
    rel_day = RelativeDay("today")
    with pytest.raises(AttributeError):
        interval = rel_day.rollforward(datetime.datetime(2022, 5, 15))


def test_repr():
    rel_day = RelativeDay("today")
    repr(rel_day)



from rocketry.parse import parse_time
from rocketry.time import TimeOfDay, TimeOfHour, TimeOfMinute, TimeDelta, TimeOfWeek, TimeOfYear, TimeOfMonth

import pytest

@pytest.mark.parametrize(
    "time_str,expected", [
        pytest.param("every 10 seconds", TimeDelta("10 seconds"), id="every"),

        pytest.param("time of minute between 15:00 and 30:00", TimeOfMinute("15:00", "30:00"), id="TimeOfMinute between"),
        pytest.param("time of minute before 30:00", TimeOfMinute(None, "30:00"), id="TimeOfMinute before"),
        pytest.param("time of minute after 30:00", TimeOfMinute("30:00", None), id="TimeOfMinute after"),

        pytest.param("time of hour between 15:00 and 30:00", TimeOfHour("15:00", "30:00"), id="TimeOfHour between"),
        pytest.param("time of hour before 30:00", TimeOfHour(None, "30:00"), id="TimeOfHour before"),
        pytest.param("time of hour after 30:00", TimeOfHour("30:00", None), id="TimeOfHour after"),

        pytest.param("time of day between 10:00 and 12:00", TimeOfDay("10:00", "12:00"), id="TimeOfDay between"),
        pytest.param("time of day before 12:00", TimeOfDay(None, "12:00"), id="TimeOfDay before"),
        pytest.param("time of day after 12:00", TimeOfDay("12:00", None), id="TimeOfDay after"),

        pytest.param("time of week on Monday", TimeOfWeek("Mon", "Mon"), id="TimeOfWeek on"),

        pytest.param("time of month between 1st and 2nd", TimeOfMonth("1st", "2nd"), id="TimeOfMonth between (1st, 2nd)"),
        pytest.param("time of month between 3rd and 4th", TimeOfMonth("3rd", "4th"), id="TimeOfMonth between (3rd, 4th)"),
    ]
)
def test_string(time_str, expected):
    cond = parse_time(time_str)
    assert cond == expected


from .interval import *
from rocketry.core.time import TimeDelta, StaticInterval, All, Any

from .construct import get_between, get_before, get_after, get_full_cycle, get_on
from .delta import TimeSpanDelta

from rocketry.session import Session

Session._time_parsers.update(
    {
        re.compile(r"time of (?P<type_>month|week|day|hour|minute) between (?P<start>.+) and (?P<end>.+)"): get_between,
        re.compile(r"time of (?P<type_>month|week|day|hour|minute) after (?P<start>.+)"): get_after,
        re.compile(r"time of (?P<type_>month|week|day|hour|minute) before (?P<end>.+)"): get_before,
        re.compile(r"time of (?P<type_>month|week) on (?P<start>.+)"): get_on,

        re.compile(r"every (?P<past>.+)"): TimeDelta,
        re.compile(r"past (?P<past>.+)"): TimeDelta,
        "always": StaticInterval(),
        "never": StaticInterval(start=StaticInterval.max - StaticInterval.resolution),
    }
)



import pytest

from rocketry.conditions import (
    ParamExists, IsPeriod,
    Any, All
)
from rocketry.conds import true, false
from rocketry.time import TimeDelta

def test_true():
    assert bool(true)

def test_false():
    assert not bool(false)

@pytest.mark.parametrize(
    "params,make_cond,expected",
    [
        pytest.param(
            {"mode": "test", "state": "right"},
            lambda: ParamExists(mode="test", state="right"),
            True, id="Dict, true"
        ),
        pytest.param(
            {"mode": "test", "state": "right"},
            lambda: ParamExists(mode="test", state="wrong"),
            False, id="Dict, false (wrong value)"
        ),
        pytest.param(
            {"mode": "test", "state": "right"},
            lambda: ParamExists(mode="test", stuff="right"),
            False, id="Dict, false (missing key)"
        ),
        pytest.param(
            {"mode": "test", "state": "right"},
            lambda: ParamExists("mode"),
            True, id="Tuple, true"
        ),
        pytest.param(
            {"mode": "test", "state": "right"},
            lambda: ParamExists("stuff"),
            False, id="Tuple, false (missing key)"
        ),
        pytest.param(
            {"mode": "test", "feel": "good", "state": "right", "distance": "short", "world": "this"},
            lambda: ParamExists("mode", "feel", state="right", distance="short"),
            True, id="Multi, true"
        ),
        pytest.param(
            {},
            lambda: ParamExists("mode", "feel", state="right", distance="short"),
            False, id="Multi, false (empty)"
        ),
    ]
)
def test_params(session, params, make_cond, expected):
    
    session.parameters.update(params)
    cond = make_cond()
    if expected:
        assert bool(cond)
    else:
        assert not bool(cond)

def test_parameter_exists(session):
    cond = ParamExists(x="yes", y="yes")
    assert not bool(cond)

    session.parameters["x"] = "yes"
    assert not bool(cond)

    session.parameters["y"] = "yes"
    assert bool(cond)

    session.parameters["y"] = "no"
    assert not bool(cond)


@pytest.mark.parametrize("get_cond,exc",
    [
        pytest.param(lambda: IsPeriod(TimeDelta("2 hours")), AttributeError, id="IsPeriod with timedelta")
    ]
)
def test_fail(get_cond,exc):
    with pytest.raises(exc):
        get_cond()

@pytest.mark.parametrize("obj,string,represent", 
    [
        pytest.param(
            All(true, false), 
            "(true & false)", 
            "All(true, false)", 
            id="All"),
        pytest.param(
            Any(true, false), 
            "(true | false)", 
            "Any(true, false)", 
            id="Any")
    ]
)
def test_representation(obj, string, represent):
    assert str(obj) == string
    assert repr(obj) == represent



import pytest
from rocketry.conditions.task.task import TaskStarted
from rocketry.core.condition.base import BaseCondition
from rocketry.tasks import FuncTask
from rocketry.exc import TaskInactionException
from rocketry.conditions import SchedulerCycles

from rocketry.conditions import FuncCond
from rocketry.conds import true

def run_succeeding():
    pass

@pytest.mark.parametrize("execution", ["main", "thread", "process"])
def test_func_cond_decorator(session, execution):

    @FuncCond(decor_return_func=False)
    def is_true():
        return True
    
    assert isinstance(is_true, BaseCondition)

    task_success = FuncTask(
        run_succeeding, 
        start_cond=true & is_true, 
        name="task success",
        execution=execution
    )

    session.config.shut_cond = TaskStarted(task=task_success) >= 1
    session.start()
    
    n_success = task_success.logger.filter_by(action="run").count()
    assert n_success == 1



import pytest
import rocketry
from rocketry.args import FuncArg

from rocketry.core import Parameters
from rocketry.args import Private
from rocketry.parameters import FuncParam

@pytest.mark.parametrize(
    "get_param,mater,repr",
    [
        pytest.param(
            lambda: Parameters({"x": 0, "y": 1}),
            {"x": 0, "y": 1},
            {"x": 0, "y": 1},
            id="Dict"),
        pytest.param(
            lambda: Parameters(x=0, y=1),
            {"x": 0, "y": 1},
            {"x": 0, "y": 1},
            id="Kwargs"),
        pytest.param(
            lambda: Parameters(password="pwd1234", user_id="myself", type_=Private),
            {"password": "pwd1234", "user_id": "myself"},
            {"password": "*****", "user_id": "*****"},
            id="Dict as private"),
    ],
)
def test_contruct(get_param, mater, repr):
    params = get_param()
    assert mater == params.materialize(task="mytask")
    assert repr == params.materialize()
    assert repr == dict(**params)


def test_from_func(session):

    assert "myparam" not in session.parameters
    assert "a_param" not in session.parameters

    @session.parameters.param_func
    def myparam():
        return 5

    assert "myparam" in session.parameters
    assert session.parameters.materialize() == {"myparam": 5}
    session.parameters.clear()

    @session.parameters.param_func(key="a_param")
    def myparam2(session):
        assert "a_param" in session.parameters.keys()
        return 5

    assert "a_param" in session.parameters
    assert session.parameters.materialize() == {"a_param": 5}

def test_func_param(session:rocketry.Session):
    @FuncParam()
    def my_param():
        return 5
    
    assert my_param() == 5
    assert 'my_param' in session.parameters
    assert isinstance(session.parameters._params['my_param'], FuncArg)
    assert session.parameters['my_param'] == 5
    assert session.parameters._params['my_param'].func is my_param

def test_func_param_named(session:rocketry.Session):
    @FuncParam(name="a_param")
    def my_param():
        return 5
    
    assert my_param() == 5
    assert 'a_param' in session.parameters
    assert isinstance(session.parameters._params['a_param'], FuncArg)
    assert session.parameters['a_param'] == 5
    assert session.parameters._params['a_param'].func is my_param



import pytest
from datetime import datetime

from rocketry.time.interval import (
    TimeOfWeek
)

# TimeOfWeek
# Year 2024 was chosen as it starts on monday
@pytest.mark.parametrize(
    "dt,start,end",
    [
        # Regular
        pytest.param(
            datetime(2024, 1, 2, 10, 00),
            "Tue 10:00", "Sat 12:00",
            id="Left of interval"),
        pytest.param(
            datetime(2024, 1, 6, 12, 00),
            "Tue 10:00", "Sat 12:00",
            id="Right of interval"),
        pytest.param(
            datetime(2024, 1, 4, 11, 00),
            "Tue 10:00", "Sat 12:00",
            id="Middle of interval"),

        # Over weekend
        pytest.param(
            datetime(2024, 1, 6, 10, 00),
            "Sat 10:00", "Tue 12:00",
            id="Left of over weekend interval"),
        pytest.param(
            datetime(2024, 1, 9, 12, 00),
            "Sat 10:00", "Tue 12:00",
            id="Right of over weekend interval"),
        pytest.param(
            datetime(2024, 1, 7, 23, 59, 59, 999999),
            "Sat 10:00", "Tue 12:00",
            id="Middle left of over weekend interval"),
        pytest.param(
            datetime(2024, 1, 8, 00, 00),
            "Sat 10:00", "Tue 12:00",
            id="Middle right of over weekend interval"),

        # Full Cycle
        pytest.param(
            datetime(2024, 1, 1, 00, 00),
            None, None,
            id="Full interval"),
        pytest.param(
            datetime(2024, 1, 2, 00, 00),
            "Tue", "Tue",
            id="Joint of full interval"),
        pytest.param(
            datetime(2024, 1, 1, 00, 00),
            "Tue 00:00", "Tue 00:00",
            id="Right of full interval"),
        pytest.param(
            datetime(2024, 1, 6, 00, 00),
            "Tue 00:00", "Tue 00:00",
            id="Left of full interval"),
    ],
)
def test_in(start, end, dt):
    time = TimeOfWeek(start, end)
    assert dt in time


@pytest.mark.parametrize(
    "dt,start,end",
    [
        # Regular
        pytest.param(
            datetime(2024, 1, 1, 0, 0),
            "Tue 10:00", "Sat 12:00",
            id="Left from interval"),
        pytest.param(
            datetime(2024, 1, 7, 14, 00, 00, 1),
            "Tue 10:00", "Sat 12:00",
            id="Right from interval"),

        # Over weekend
        pytest.param(
            datetime(2024, 1, 6, 9, 59, 59, 999999),
            "Sat 10:00", "Tue 12:00",
            id="Left from over weekend interval"),
        pytest.param(
            datetime(2024, 1, 9, 12, 1, 00),
            "Sat 10:00", "Tue 12:00",
            id="Right from over weekend interval"),
    ],
)
def test_not_in(start, end, dt):
    time = TimeOfWeek(start, end)
    assert dt not in time



import asyncio
import datetime
import logging
import time
import os, re
import multiprocessing

import pytest
from redbird.logging import RepoHandler
from redbird.repos import MemoryRepo

from rocketry.log.log_record import LogRecord, TaskLogRecord
import rocketry
from rocketry import Session
from rocketry.core import Scheduler, Parameters
from rocketry.log.log_record import MinimalRecord
from rocketry.tasks import FuncTask
from rocketry.time import TimeDelta
from rocketry.exc import TaskInactionException, TaskTerminationException
from rocketry.conditions import SchedulerCycles, SchedulerStarted, TaskStarted, AlwaysFalse, AlwaysTrue
from rocketry.args import Private, TerminationFlag

from rocketry.conds import true, false

def run_failing():
    raise RuntimeError("Task failed")

def run_succeeding():
    pass

def run_inacting():
    raise TaskInactionException()

async def run_failing_async():
    raise RuntimeError("Task failed")

async def run_succeeding_async():
    pass

async def run_inacting_async():
    raise TaskInactionException()


def run_slow():
    time.sleep(5)

def run_slow_thread(flag=TerminationFlag()):
    t = 0
    while not flag.is_set() and t < 5:
        time.sleep(0.001)
        t += 0.001
    if flag.is_set():
        raise TaskTerminationException()

async def run_slow_async():
    await asyncio.sleep(5)


def create_line_to_file():
    with open("work.txt", "a") as file:
        file.write("line created\n")

async def create_line_to_file_async():
    with open("work.txt", "a") as file:
        file.write("line created\n")

def run_with_param(int_5):
    assert int_5 == 5

def run_creating_child():

    proc = multiprocessing.Process(target=run_succeeding, daemon=True)
    proc.start()

def test_scheduler_shut_cond(session):
    assert not session.scheduler.check_shut_cond(None)
    
    assert session.scheduler.check_shut_cond(true)
    assert not session.scheduler.check_shut_cond(~true)
    assert session.scheduler.check_shut_cond(true & true)
    assert not session.scheduler.check_shut_cond(false)

@pytest.mark.parametrize("execution", ["main", "async", "thread", "process"])
@pytest.mark.parametrize("func", [pytest.param(create_line_to_file, id="sync"), pytest.param(create_line_to_file_async, id="async")])
def test_task_execution(tmpdir, execution, func, session):
    with tmpdir.as_cwd() as old_dir:
        # To be confident the scheduler won't lie to us
        # we test the task execution with a job that has
        # actual measurable impact outside rocketry
        FuncTask(func, name="add line to file", start_cond=AlwaysTrue(), execution=execution),

        session.config.shut_cond = (TaskStarted(task="add line to file") >= 3) | ~SchedulerStarted(period=TimeDelta("5 second"))

        session.start()
        # Sometimes in CI the task may end up to be started only twice thus we tolerate slightly
        with open("work.txt", "r") as file:
            assert 2 <= len(list(file))

@pytest.mark.parametrize("get_handler", [
    pytest.param(lambda: RepoHandler(repo=MemoryRepo(model=TaskLogRecord)), id="Memory with model"),
    pytest.param(lambda: RepoHandler(repo=MemoryRepo()), id="Memory with dict"),
])
@pytest.mark.parametrize("execution", ["main", "thread", "process"])
@pytest.mark.parametrize(
    "task_func,run_count,fail_count,success_count,inact_count",
    [
        pytest.param(
            run_succeeding, 
            3, 0, 3, 0,
            id="Succeeding task"),

        pytest.param(
            run_failing, 
            3, 3, 0, 0,
            id="Failing task"),
        pytest.param(
            run_inacting, 
            3, 0, 0, 3,
            id="Inacting task"),
    ],
)
def test_task_log(tmpdir, execution, task_func, run_count, fail_count, success_count, inact_count, get_handler):
    """Test the task logging thoroughly including the common
    logging schemes, execution (main, thread, process) and 
    outcomes (fail, success, inaction).

    This task may take some time but should cover most of the
    issues with logging.
    """

    with tmpdir.as_cwd() as old_dir:
        # Set session (and logging)
        session = Session(config={"debug": True, "silence_task_prerun": False})
        rocketry.session = session
        session.set_as_default()

        task_logger = logging.getLogger(session.config.task_logger_basename)
        task_logger.handlers = [
            get_handler()
        ]

        task = FuncTask(task_func, name="mytask", start_cond=AlwaysTrue(), execution=execution)

        session.config.shut_cond = (TaskStarted(task="mytask") >= run_count) | ~SchedulerStarted(period=TimeDelta("10 second"))
        session.start()

        assert (TaskStarted(task="mytask") >= run_count).observe(session=session)

        # Test history
        history = list(task.logger.get_records())
        logger = task.logger
        assert run_count == logger.filter_by(action="run").count()
        assert success_count == logger.filter_by(action="success").count()
        assert fail_count == logger.filter_by(action="fail").count()
        assert inact_count == logger.filter_by(action="inaction").count()

        # Test relevant log items
        for record in history:
            if not isinstance(record, dict):
                record = record.dict()
            assert record["task_name"] == "mytask"
            assert isinstance(record["created"], float)
            assert isinstance(record["start"], datetime.datetime)
            if record["action"] != "run":
                assert isinstance(record["end"], datetime.datetime)
                assert isinstance(record["runtime"], datetime.timedelta)

            # Test traceback
            if record["action"] == "fail":
                assert re.match(
                    r'Traceback \(most recent call last\):\n  File ".+", line [0-9]+, in [\s\S]+, in run_failing\n    raise RuntimeError\("Task failed"\)\nRuntimeError: Task failed', 
                    record["exc_text"]
                )

        # Test some other relevant APIs
        assert history == list(task.logger.get_records())
        assert history[-1] == task.logger.get_latest()
        assert history == list(session.get_task_log())

        assert run_count == len(list(task.logger.get_records(action="run")))
        assert success_count == len(list(task.logger.get_records(action="success")))
        assert fail_count == len(list(task.logger.get_records(action="fail")))
        assert inact_count == len(list(task.logger.get_records(action="inaction")))

@pytest.mark.parametrize("mode", ["use logs", "use cache"])
@pytest.mark.parametrize("func_type", ["sync", "async"])
@pytest.mark.parametrize("execution", ["main", "thread", "process"])
def test_task_status(session, execution, func_type, mode):
    session.config.force_status_from_logs = True if mode == "use logs" else False

    task_success = FuncTask(
        run_succeeding if func_type == "sync" else run_succeeding_async, 
        start_cond=true, 
        name="task success",
        execution=execution,
        priority=0
    )
    task_fail = FuncTask(
        run_failing if func_type == "sync" else run_failing_async, 
        start_cond=true, 
        name="task fail",
        execution=execution,
        priority=0
    )
    task_inact = FuncTask(
        run_inacting if func_type == "sync" else run_inacting_async, 
        start_cond=true, 
        name="task inact",
        execution=execution,
        priority=100
    )
    task_not_run = FuncTask(
        run_inacting if func_type == "sync" else run_inacting_async, 
        start_cond=false, 
        name="task not run",
        execution=execution,
        priority=0
    )
    session.config.shut_cond = (TaskStarted(task="task inact") >= 3) | ~SchedulerStarted(period=TimeDelta("20 second"))
    session.start()

    # Test status
    assert task_success.last_run is not None
    assert task_success.last_success is not None
    assert task_success.last_fail is None

    assert task_fail.last_run is not None
    assert task_fail.last_fail is not None
    assert task_fail.last_success is None

    assert task_inact.last_run is not None

    assert task_not_run.last_run is None
    assert task_not_run.last_success is None
    assert task_not_run.last_fail is None

    assert task_success.status == "success"
    assert task_fail.status == "fail"
    assert task_inact.status == "inaction"
    assert task_not_run.status == None

    # Test logs
    assert 3 == task_success.logger.filter_by(action="run").count()
    assert 3 == task_fail.logger.filter_by(action="run").count()
    assert 3 == task_inact.logger.filter_by(action="run").count()
    assert 0 == task_not_run.logger.filter_by(action="run").count()

    assert 3 == task_success.logger.filter_by(action="success").count()
    assert 0 == task_fail.logger.filter_by(action="success").count()
    assert 0 == task_inact.logger.filter_by(action="suceess").count()

    assert 0 == task_success.logger.filter_by(action="fail").count()
    assert 3 == task_fail.logger.filter_by(action="fail").count()
    assert 0 == task_inact.logger.filter_by(action="fail").count()

    assert 0 == task_success.logger.filter_by(action="inaction").count()
    assert 0 == task_fail.logger.filter_by(action="inaction").count()
    assert 3 == task_inact.logger.filter_by(action="inaction").count()

@pytest.mark.parametrize("execution", ["main", "thread", "process"])
def test_task_force_run(tmpdir, execution, session):
    with tmpdir.as_cwd() as old_dir:
        task = FuncTask(
            run_succeeding, 
            start_cond=AlwaysFalse(), 
            name="task",
            execution=execution
        )
        task.force_run = True

        session.config.shut_cond = SchedulerCycles() >= 5
        session.start()

        logger = task.logger
        assert 1 == logger.filter_by(action="run").count()

        # The force_run should have reseted as it should have run once
        assert not task.force_run


@pytest.mark.parametrize("execution", ["main", "thread", "process"])
def test_task_disabled(tmpdir, execution, session):
    with tmpdir.as_cwd() as old_dir:

        task = FuncTask(
            run_succeeding, 
            start_cond=AlwaysFalse(), 
            name="task",
            execution=execution
        )
        task.disabled = True

        session.config.shut_cond = SchedulerCycles() >= 5
        session.start()

        history = task.logger.get_records()
        assert 0 == sum([record for record in history if record["action"] == "run"])

        assert task.disabled


@pytest.mark.parametrize("execution", ["main", "thread", "process"])
def test_task_force_disabled(tmpdir, execution, session):
    # NOTE: force_run overrides disabled
    # as it is more practical to keep 
    # a task disabled and force it running
    # manually than prevent force run with
    # disabling
    with tmpdir.as_cwd() as old_dir:

        task = FuncTask(
            run_succeeding, 
            start_cond=AlwaysFalse(), 
            name="task",
            execution=execution
        )
        task.disabled = True
        task.force_run = True

        session.config.shut_cond = SchedulerCycles() >= 5
        session.start()

        logger = task.logger
        assert 1 == logger.filter_by(action="run").count()
        assert 1 == logger.filter_by(action="success").count()

        assert task.disabled
        assert not task.force_run # This should be reseted

@pytest.mark.parametrize("execution", ["main", "thread", "process"])
def test_priority(tmpdir, execution, session):
    with tmpdir.as_cwd() as old_dir:

        task_1 = FuncTask(run_succeeding, name="1", priority=100, start_cond=AlwaysTrue(), execution=execution)
        task_3 = FuncTask(run_failing, name="3", priority=10, start_cond=AlwaysTrue(), execution=execution)
        task_2 = FuncTask(run_failing, name="2", priority=50, start_cond=AlwaysTrue(), execution=execution)
        task_4 = FuncTask(run_failing, name="4", start_cond=AlwaysTrue(), execution=execution)

        assert 0 == task_4.priority

        session.config.shut_cond = (SchedulerCycles() == 1) | ~SchedulerStarted(period=TimeDelta("2 seconds"))

        session.start()
        assert session.scheduler.n_cycles == 1 

        task_1_start = list(task_1.logger.get_records())[0].created
        task_2_start = list(task_2.logger.get_records())[0].created
        task_3_start = list(task_3.logger.get_records())[0].created
        task_4_start = list(task_4.logger.get_records())[0].created
        
        assert task_1_start < task_2_start < task_3_start < task_4_start

@pytest.mark.parametrize("execution", ["main", "thread", "process"])
def test_pass_params_as_global(tmpdir, execution, session):
    # thread-Parameters has been observed to fail rarely
    with tmpdir.as_cwd() as old_dir:

        task = FuncTask(run_with_param, name="parametrized", start_cond=AlwaysTrue(), execution=execution)

        session.config.shut_cond = (TaskStarted(task="parametrized") >= 1) | ~SchedulerStarted(period=TimeDelta("2 seconds"))

        # Passing global parameters
        session.parameters["int_5"] = 5
        session.parameters["extra_param"] = "something"

        session.scheduler()

        logger = task.logger
        assert 1 == logger.filter_by(action="run").count()
        assert 1 == logger.filter_by(action="success").count()
        assert 0 == logger.filter_by(action="fail").count()

@pytest.mark.parametrize("parameters", [
    pytest.param({"int_5": 5}, id="dict"), 
    pytest.param(Parameters(int_5=5), id="Parameters"),
    pytest.param(Parameters(int_5=Private(5)), id="Parameter with secret"),
])
@pytest.mark.parametrize("execution", ["main", "thread", "process"])
def test_pass_params_as_local(tmpdir, execution, parameters, session):
    with tmpdir.as_cwd() as old_dir:

        task = FuncTask(
            run_with_param, 
            name="parametrized", 
            parameters=parameters,
            start_cond=AlwaysTrue(),
            execution=execution
        )
        session.config.shut_cond = (TaskStarted(task="parametrized") >= 1) | ~SchedulerStarted(period=TimeDelta("2 seconds"))

        session.start()

        logger = task.logger
        assert 1 == logger.filter_by(action="run").count()
        assert 1 == logger.filter_by(action="success").count()
        assert 0 == logger.filter_by(action="fail").count()

@pytest.mark.parametrize("execution", ["main", "thread", "process"])
def test_pass_params_as_local_and_global(tmpdir, execution, session):
    with tmpdir.as_cwd() as old_dir:

        task = FuncTask(
            run_with_param, 
            name="parametrized", 
            parameters={"int_5": 5},
            start_cond=AlwaysTrue(),
            execution=execution
        )

        session.config.shut_cond = (TaskStarted(task="parametrized") >= 1) | ~SchedulerStarted(period=TimeDelta("2 seconds"))

        # Additional parameters
        session.parameters["extra_param"] = "something"

        session.start()

        logger = task.logger
        assert 1 == logger.filter_by(action="run").count()
        assert 1 == logger.filter_by(action="success").count()
        assert 0 == logger.filter_by(action="fail").count()


# Only needed for testing start up and shutdown
def create_line_to_startup_file():
    with open("start.txt", "w") as file:
        file.write("line created\n")

def create_line_to_shutdown():
    with open("shut.txt", "w") as file:
        file.write("line created\n")

@pytest.mark.parametrize("execution", ["main", "thread", "process"])
def test_startup_shutdown(tmpdir, execution, session):
    with tmpdir.as_cwd() as old_dir:
        
        FuncTask(create_line_to_startup_file, name="startup", on_startup=True, execution=execution)
        FuncTask(create_line_to_shutdown, name="shutdown", on_shutdown=True, execution=execution)

        session.config.shut_cond = AlwaysTrue()

        session.start()
        if execution == "thread":
            # It may take a moment for the
            # thread to write to the file
            timeout = 5
            time_taken = 0
            while os.path.exists("shut.txt"):
                time.sleep(0.01)
                time_taken += 0.01
                if time_taken > timeout:
                    break
        
        assert os.path.exists("start.txt")
        assert os.path.exists("shut.txt")

        assert list(session.get_task_log()) != []

@pytest.mark.parametrize("execution", ["main", "thread", "process"])
def test_logging_repo(tmpdir, execution):
    from redbird.logging import RepoHandler
    from redbird.repos import MemoryRepo
    session = Session()
    session.set_as_default()

    handler = RepoHandler(repo=MemoryRepo(model=MinimalRecord))

    logger = logging.getLogger("rocketry.task")
    logger.handlers = []
    logger.addHandler(handler)


    with tmpdir.as_cwd() as old_dir:

        task_1 = FuncTask(run_succeeding, name="1", priority=100, start_cond=AlwaysTrue(), execution=execution)
        task_3 = FuncTask(run_failing, name="3", priority=10, start_cond=AlwaysTrue(), execution=execution)
        task_2 = FuncTask(run_failing, name="2", priority=50, start_cond=AlwaysTrue(), execution=execution)
        task_4 = FuncTask(run_failing, name="4", start_cond=AlwaysTrue(), execution=execution)

        assert 0 == task_4.priority

        session.config.shut_cond = (SchedulerCycles() == 1) | ~SchedulerStarted(period=TimeDelta("2 seconds"))
        session.start()
        assert session.scheduler.n_cycles == 1 

        task_1_start = list(task_1.logger.get_records())[0].created
        task_2_start = list(task_2.logger.get_records())[0].created
        task_3_start = list(task_3.logger.get_records())[0].created
        task_4_start = list(task_4.logger.get_records())[0].created
        
        assert task_1_start < task_2_start < task_3_start < task_4_start

@pytest.mark.parametrize("execution", ["async", "thread", "process"])
def test_instant_shutdown(execution, session):
    assert not session.config.instant_shutdown
    session.config.instant_shutdown = True

    func = {"async": run_slow_async, "thread": run_slow_thread, "process": run_slow}[execution]

    task = FuncTask(func, execution=execution, start_cond=true, end_cond=SchedulerCycles() == 2)

    session.config.shut_cond = SchedulerCycles() == 2

    session.start()

    assert 1 == task.logger.filter_by(action="run").count()
    assert 2 == task.logger.filter_by().count()

    assert 0 == task.logger.filter_by(action="fail").count()
    assert 0 == task.logger.filter_by(action="success").count()
    assert 1 == task.logger.filter_by(action="terminate").count()



import itertools

import pytest

from rocketry.parse.utils import ParserError
from rocketry.conditions.scheduler import SchedulerCycles, SchedulerStarted
from rocketry.parse.condition import parse_condition
from rocketry.conditions import ParamExists
from rocketry.conditions import (
    AlwaysTrue, AlwaysFalse, 
    All, Any, Not,

    TaskRunning, TaskStarted, 
    TaskFailed, TaskSucceeded, TaskFinished,
    TaskInacted, TaskTerminated,
    DependFailure, DependSuccess, DependFinish,
    TaskExecutable,

    IsPeriod,

    IsEnv,
)

from rocketry.time import (
    TimeDelta,

    TimeOfMinute,
    TimeOfHour,
    TimeOfDay,
    TimeOfWeek,
    TimeOfMonth
)

from rocketry.conds import (
    minutely, hourly, daily, weekly, monthly
)

cases_time = [
    pytest.param("minutely", minutely, id="hourly"),
    pytest.param("hourly", hourly, id="hourly"),
    pytest.param("daily", daily, id="daily"),
    pytest.param("weekly", weekly, id="weekly"),
    pytest.param("monthly", monthly, id="monthly"),

    pytest.param("minutely starting 45:00",   TaskExecutable(period=TimeOfMinute("45:00", "45:00")), id="hourly starting"),
    pytest.param("hourly starting 45:00",   TaskExecutable(period=TimeOfHour("45:00", "45:00")), id="hourly starting"),
    pytest.param("daily starting 10:00",    TaskExecutable(period=TimeOfDay("10:00", "10:00")),  id="daily starting"),
    pytest.param("weekly starting Tuesday", TaskExecutable(period=TimeOfWeek("Tue", "Tue")),     id="weekly starting"),
    pytest.param("monthly starting 1.",  TaskExecutable(period=TimeOfMonth("1.", "1.")), id="monthly starting"),

    pytest.param("hourly between 45:00 and 50:00",       TaskExecutable(period=TimeOfHour("45:00", "50:00")), id="hourly between"),
    pytest.param("daily between 10:00 and 14:00",        TaskExecutable(period=TimeOfDay("10:00", "14:00")),  id="daily between"),
    pytest.param("weekly between Tuesday and Wednesday", TaskExecutable(period=TimeOfWeek("Tue", "Wed")),     id="weekly between"),
    pytest.param("monthly between 1. and 15.",  TaskExecutable(period=TimeOfMonth("1.", "15.")), id="monthly between"),

    pytest.param("hourly after 45:00",   TaskExecutable(period=TimeOfHour("45:00", None)), id="hourly after"),
    pytest.param("daily after 10:00",    TaskExecutable(period=TimeOfDay("10:00", None)),  id="daily after"),
    pytest.param("weekly after Tuesday", TaskExecutable(period=TimeOfWeek("Tue", None)),   id="weekly after"),
    pytest.param("monthly after 1.",  TaskExecutable(period=TimeOfMonth("1.", None)), id="monthly after"),

    pytest.param("hourly before 45:00",   TaskExecutable(period=TimeOfHour(None, "45:00")), id="hourly before"),
    pytest.param("daily before 10:00",    TaskExecutable(period=TimeOfDay(None, "10:00")), id="daily before"),
    pytest.param("weekly before Tuesday", TaskExecutable(period=TimeOfWeek(None, "Tue")), id="weekly before"),
    pytest.param("monthly before 1.",  TaskExecutable(period=TimeOfMonth(None, "1.")), id="monthly before"),

    pytest.param("weekly on Tuesday", TaskExecutable(period=TimeOfWeek("Tue", time_point=True)), id="weekly on"),

    # Time delta
    pytest.param("every 1 hours", TaskStarted(period=TimeDelta(past="1 hours")) == 0, id="every hour"),
    pytest.param("every 1 days 1 hours 30 minutes 30 seconds", TaskStarted(period=TimeDelta(past="1 days 1 hours 30 minutes 30 seconds")) == 0, id="every hour 30 mins 30 seconds"),

    # IsTimeOf...
    pytest.param("time of hour between 45:00 and 50:00",       IsPeriod(period=TimeOfHour("45:00", "50:00")), id="time of hour between"),
    pytest.param("time of day between 10:00 and 14:00",        IsPeriod(period=TimeOfDay("10:00", "14:00")), id="time of day between"),
    pytest.param("time of week between Tuesday and Wednesday", IsPeriod(period=TimeOfWeek("Tue", "Wed")), id="time of week between"),
    pytest.param("time of month between 1st and 2nd",  IsPeriod(period=TimeOfMonth("1st", "2nd")), id="time of month between (1st, 2nd)"),

    pytest.param("time of hour after 45:00",   IsPeriod(period=TimeOfHour("45:00", None)), id="time of hour after"),
    pytest.param("time of day after 10:00",    IsPeriod(period=TimeOfDay("10:00", None)), id="time of day after"),
    pytest.param("time of week after Tuesday", IsPeriod(period=TimeOfWeek("Tue", None)), id="time of week after"),
    pytest.param("time of month after 1.",  IsPeriod(period=TimeOfMonth("1.", None)), id="time of month after"),

    pytest.param("time of hour before 45:00",   IsPeriod(period=TimeOfHour(None, "45:00")), id="time of hour before"),
    pytest.param("time of day before 10:00",    IsPeriod(period=TimeOfDay(None, "10:00")), id="time of day before"),
    pytest.param("time of week before Tuesday", IsPeriod(period=TimeOfWeek(None, "Tue")), id="time of week before"),
    pytest.param("time of week on Tuesday", IsPeriod(period=TimeOfWeek("Tue", "Tue")), id="time of week on"),
    pytest.param("time of month before 1.",  IsPeriod(period=TimeOfMonth(None, "1.")), id="time of month before"),
]

cases_task = [
    # Single task related
    pytest.param("task 'mytask' is running", TaskRunning(task="mytask"), id="while task running"),

    # Task dependent related (requires 2 tasks)
    pytest.param("after task 'other'",           DependSuccess(depend_task="other"), id="after task"),
    pytest.param("after task 'other' succeeded", DependSuccess(depend_task="other"), id="after task success"),
    pytest.param("after task 'other' failed",    DependFailure(depend_task="other"), id="after failed"),
    pytest.param("after task 'other' finished",  DependFinish(depend_task="other"), id="after finished"),
    pytest.param("after task 'group1.group-2.mytask+'", DependSuccess(depend_task="group1.group-2.mytask+"), id="after task special chars"),

    # Multi
    pytest.param("after tasks 'first', 'second', 'third'", All(DependSuccess(depend_task="first"), DependSuccess(depend_task="second"), DependSuccess(depend_task="third")), id="after task (multi)"),
    pytest.param("after tasks 'first', 'second', 'third' succeeded", All(DependSuccess(depend_task="first"), DependSuccess(depend_task="second"), DependSuccess(depend_task="third")), id="after task success (multi)"),
    pytest.param("after tasks 'first', 'second', 'third' failed", All(DependFailure(depend_task="first"), DependFailure(depend_task="second"), DependFailure(depend_task="third")), id="after task fail (multi)"),
    pytest.param("after tasks 'first', 'second', 'third' finished", All(DependFinish(depend_task="first"), DependFinish(depend_task="second"), DependFinish(depend_task="third")), id="after task fail (multi)"),

    pytest.param("after any tasks 'first', 'second', 'third' succeeded", Any(DependSuccess(depend_task="first"), DependSuccess(depend_task="second"), DependSuccess(depend_task="third")), id="after task success (multi)"),
    pytest.param("after any tasks 'first', 'second', 'third' failed", Any(DependFailure(depend_task="first"), DependFailure(depend_task="second"), DependFailure(depend_task="third")), id="after task fail (multi)"),
    pytest.param("after any tasks 'first', 'second', 'third' finished", Any(DependFinish(depend_task="first"), DependFinish(depend_task="second"), DependFinish(depend_task="third")), id="after task fail (multi)"),


    pytest.param("has failed today", TaskFailed(period=TimeOfDay()), id="has failed today"),
]

cases_scheduler = [
    pytest.param("scheduler has more than 3 cycles", SchedulerCycles() > 3, id="scheduler cycles greater than"),
    pytest.param("scheduler has less than 3 cycles", SchedulerCycles() < 3, id="scheduler cycles less than"),
    pytest.param("scheduler started 20 minutes ago", SchedulerStarted(period=TimeDelta("20 minutes")), id="scheduler started in"),
    pytest.param("scheduler has run over 20 minutes", Not(SchedulerStarted(period=TimeDelta("20 minutes"))), id="scheduler run over"),
]

for (cls, action), (str_time, period) in itertools.product(
    [
        (TaskSucceeded, "succeeded"), 
        (TaskFailed, "failed"), 
        (TaskTerminated, "terminated"), 
        (TaskInacted, "inacted"), 
        (TaskStarted, "started")
    ], 
    [
        ("today between 10:00 and 12:00", TimeOfDay("10:00", "12:00")),
        ("this week between Mon and Fri", TimeOfWeek("Mon", "Fri")),
        ("this month between 1st and 5th", TimeOfMonth("1st", "5th")),

        ("today before 10:00", TimeOfDay(None, "10:00")),
        ("today", TimeOfDay()),
        ('', None), # Testing with no time period
        ("in past 15 minutes", TimeDelta("15 min")),
        ("past 15 minutes", TimeDelta("15 min")),
    ],
):
    # This is cryptic but just generates bunch of test cases quickly
    # for different condition classes with different time objects.
    # See ids of genereted tests for what the strings are that are 
    # tested.
    if str_time:
        str_command = f"task 'mytask' has {action} {str_time}"
    else:
        str_command = f"task 'mytask' has {action}"
    cases_task.append(pytest.param(str_command, cls(task="mytask", period=period), id=str_command))

cases_logical = [
    pytest.param("always true", AlwaysTrue(), id="AlwaysTrue"),
    pytest.param("always false", AlwaysFalse(), id="AlwaysTrue"),
    pytest.param("true", AlwaysTrue(), id="true"),
    pytest.param("false", AlwaysFalse(), id="false"),

    # Logical operations
    pytest.param("always true & always false", All(AlwaysTrue(), AlwaysFalse()), id="Logical AND"),
    pytest.param("always true | always false", Any(AlwaysTrue(), AlwaysFalse()), id="Logical OR"),
    pytest.param("~ always false", Not(AlwaysFalse()), id="Logical NOT"),
]

cases_misc = [
    pytest.param("""(true & true)
        | (false & ~
        false)""", (AlwaysTrue() & AlwaysTrue()) | (AlwaysFalse() & ~AlwaysFalse()), id="Multiline"),
    pytest.param("env 'test'", IsEnv('test'), id="IsEnv 'test'"),
    pytest.param("param 'x' exists", ParamExists('x'), id="ParamExists 'x'"),
    pytest.param("param 'x' is 'myval'", ParamExists(x='myval'), id="ParamExists 'x=5'"),
]


# All cases
cases = cases_logical + cases_task + cases_time + cases_misc + cases_scheduler

@pytest.mark.parametrize(
    "cond_str,expected", cases
)
def test_string(cond_str, expected):
    cond = parse_condition(cond_str)
    assert cond == expected

@pytest.mark.parametrize(
    "cond_str,expected", cases
)
def test_back_to_string(cond_str, expected):
    cond = parse_condition(cond_str)
    cond_as_str = str(cond)
    cond_2 = parse_condition(cond_as_str)
    assert cond == cond_2
    #assert cond_str == cond_as_str

@pytest.mark.parametrize("cond_str,exc",
    [
        pytest.param("this is not valid", ParserError, id="Invalid condition"),
        pytest.param("true &", IndexError, id="AND, missing right"),
        pytest.param("true |", IndexError, id="OR, missing right"),
        pytest.param("& true", IndexError, id="AND, missing left"),
        pytest.param("| true", IndexError, id="OR, missing left"),

        pytest.param("after tasks missing quotes", ParserError, id="Malformed after tasks"),
    ]
)
def test_failure(cond_str, exc):
    with pytest.raises(exc):
        parse_condition(cond_str)



from datetime import datetime
import pytest
from rocketry.time.interval import (
    TimeOfDay
)

@pytest.mark.parametrize(
    "dt,start,end",
    [
        # Regular
        pytest.param(
            datetime(2020, 1, 1, 10, 00),
            "10:00", "12:00",
            id="Left of interval"),
        pytest.param(
            datetime(2020, 1, 1, 12, 00),
            "10:00", "12:00",
            id="Right of interval"),
        pytest.param(
            datetime(2020, 1, 1, 11, 00),
            "10:00", "12:00",
            id="Middle of interval"),

        # Overnight
        pytest.param(
            datetime(2020, 1, 1, 22, 00),
            "22:00", "02:00",
            id="Left of overnight interval"),
        pytest.param(
            datetime(2020, 1, 1, 2, 00),
            "22:00", "02:00",
            id="Right of overnight interval"),
        pytest.param(
            datetime(2020, 1, 1, 23, 59, 59, 999999),
            "22:00", "02:00",
            id="Middle left of overnight interval"),
        pytest.param(
            datetime(2020, 1, 1, 00, 00),
            "22:00", "02:00",
            id="Middle right of overnight interval"),

        # Full Cycle
        pytest.param(
            datetime(2020, 1, 1, 10, 00),
            None, None,
            id="Full interval"),
        pytest.param(
            datetime(2020, 1, 1, 10, 00),
            "10:00", "10:00",
            id="Joint of full interval"),
        pytest.param(
            datetime(2020, 1, 1, 12, 00),
            "10:00", "10:00",
            id="Right of full interval"),
        pytest.param(
            datetime(2020, 1, 1, 8, 00),
            "10:00", "10:00",
            id="Left of full interval"),
    ],
)
def test_in(start, end, dt):
    time = TimeOfDay(start, end)
    assert dt in time


@pytest.mark.parametrize(
    "dt,start,end",
    [
        # Regular
        pytest.param(
            datetime(2020, 1, 1, 9, 59, 59, 999999),
            "10:00", "12:00",
            id="Left from interval"),
        pytest.param(
            datetime(2020, 1, 1, 12, 00, 00, 1),
            "10:00", "12:00",
            id="Right from interval"),

        # Overnight
        pytest.param(
            datetime(2020, 1, 1, 21, 59, 59, 999999),
            "22:00", "02:00",
            id="Left from overnight interval"),
        pytest.param(
            datetime(2020, 1, 1, 2, 00, 00, 1),
            "22:00", "02:00",
            id="Right from overnight interval"),
    ],
)
def test_not_in(start, end, dt):
    time = TimeOfDay(start, end)
    assert dt not in time



import datetime
from typing import Union
from dateutil.parser import parse

ABBREVIATIONS = {
    'ns': 'nanosecond',
    'μs': 'microsecond',
    'ms': 'millisecond',
    's': 'second',
    'm': 'minute',
    'h': 'hour',
}

def to_datetime(s):
    if isinstance(s, datetime.datetime):
        return s
    elif isinstance(s, str):
        return string_to_datetime(s)
    elif hasattr(s, "timestamp"):
        # Is datetime-like. Tests' monkeypatching
        # overrides datetime.datetime thus we cannot
        # always rely on type
        return datetime.datetime.fromtimestamp(s.timestamp())
    else:
        raise TypeError(f"Cannot convert to datetime: {type(s)}")

def to_timedelta(s, **kwargs):
    "Convert object to timedelta"
    if isinstance(s, datetime.timedelta):
        return s
    elif isinstance(s, str):
        return string_to_timedelta(s)
    elif isinstance(s, (int, float)):
        return numb_to_timedelta(s, **kwargs)
    else:
        raise TypeError(f"Cannot convert to timedelta: {type(s)}")


def string_to_datetime(s):
    return parse(s)


def numb_to_timedelta(n: Union[float, int], unit="ms"):
    
    if unit == "ns":
        unit = 'μs'
        n = n / 1000
    units = ABBREVIATIONS[unit] + "s"
    return datetime.timedelta(**{units: n})

def string_to_timedelta(s:str):
    "Convert string to timedelta"

    def is_numeric_char(s):
        return s.isdigit() or s in ('.',)

    def is_wordbreak_char(s):
        return s in (' ', ',', ':', ',')

    def skip_wordbreak(s):
        for pos, char in enumerate(s):
            if not is_wordbreak_char(char):
                break
        return pos

    def get_number(s):
        num = ""
        for pos, char in enumerate(s):
            if is_numeric_char(char):
                num += char
            else:
                break
        return num, pos

    def get_unit(s):
        abbr = ""
        for pos, char in enumerate(s):
            if is_numeric_char(char) or is_wordbreak_char(char):
                break
            else:
                abbr += char
        else:
            # String ended
            pos += 1
        return abbrs[abbr], pos

    def get_hhmmss(s):
        hh, mm, ss = s.split(":")
        return to_nanoseconds(hour=int(hh), minute=int(mm), second=float(ss))

    # https://github.com/pandas-dev/pandas/blob/e8093ba372f9adfe79439d90fe74b0b5b6dea9d6/pandas/_libs/tslibs/timedeltas.pyx#L296
    abbrs = {
        'millisecond': 'millisecond',
        'ms': 'millisecond',

        'seconds': 'second',
        'second': 'second',
        'sec': 'second',
        's': 'second',

        'minutes': 'minute',
        'minute': 'minute',
        'mins': 'minute',
        'min': 'minute',
        'm': 'minute',

        'hours': 'hour',
        'hour': 'hour',
        'h': 'hour',

        'days': 'day',
        'day': 'day',
        'd': 'day',
    }

    ns = 0
    # Finding out the leading "-"
    is_negative = False
    for i, char in enumerate(s):
        if char == '-':
            is_negative = not is_negative
        elif char == "+":
            continue
        else:
            break
    s = s[i:]
    # Continuing to find the time elements
    while s:
        pos = skip_wordbreak(s)
        s = s[pos:]

        # Example: "-  2.5  days ..."
        # Pos:         ^ 

        numb, pos = get_number(s)
        s = s[pos:]

        if s[0] == ":":
            # Expecting HH:MM:SS
            ns += get_hhmmss(numb + s)
            break

        # Example: "-  2.5  days ..."
        # Pos:            ^ 

        pos = skip_wordbreak(s)
        s = s[pos:]

        # Example: "-  2.5  days ..."
        # Pos:              ^ 

        abbr, pos = get_unit(s)
        s = s[pos:]

        ns += to_nanoseconds(**{abbr: float(numb)})
    
    if is_negative:
        ns = -ns
    return datetime.timedelta(microseconds=ns / 1000)

def to_nanoseconds(day=0, hour=0, minute=0, second=0, millisecond=0, microsecond=0, nanosecond=0) -> int:
    "Turn time components to nanoseconds"
    return nanosecond + microsecond * 1_000 + millisecond * 1_000_000 + second * int(1e+9) + minute * int(6e+10) + hour * int(3.6e+12) + day * int(8.64e+13)



import time
import datetime
from typing import Tuple, Union

from .base import TimePeriod
from rocketry.pybox.time import to_nanoseconds, to_timedelta

# Conversions
def to_dict(dt):
    return {
        "year": dt.year,
        "month": dt.month,
        "day": dt.day,
        "hour": dt.hour,
        "minute": dt.minute,
        "second": dt.second,
        "microsecond": dt.microsecond,
        "nanosecond": dt.nanosecond if hasattr(dt, "nanosecond") else 0
    }


def timedelta_to_dict(dt, days_in_year=365, days_in_month=30, units=None):
    
    total_seconds = dt.total_seconds()
    nanoseconds = total_seconds * 1e+9

    components = {}
    if units == "all":
        allowed_units = (
            "years", "months", "weeks", 
            "days", "hours", "minutes", 
            "seconds", "microseconds", 
            "nanoseconds")
    elif units == "fixed" or units is None:
        # Only fixed length units
        allowed_units = (
            "days", "hours", "minutes", 
            "seconds", "microseconds", 
            "nanoseconds")
    else:
        allowed_units = units

    # Non fixed units
    if "years" in allowed_units:
        components["years"] = int(nanoseconds // (8.64e+13 * days_in_year))
        nanoseconds = nanoseconds - (components["years"] * 8.64e+13 * days_in_year)
    
    if "months" in allowed_units:
        components["months"] = int(nanoseconds // (8.64e+13 * days_in_month))
        nanoseconds = nanoseconds - (components["months"] * 8.64e+13 * days_in_month)
    
    # Alternating units
    if "weeks" in allowed_units:
        components["weeks"] = int(nanoseconds // (8.64e+13 * 7))
        nanoseconds = nanoseconds - components["weeks"] * 8.64e+13 * 7
    
    # Fixed units
    if "days" in allowed_units:
        components["days"] = int(nanoseconds // 8.64e+13)
        nanoseconds = nanoseconds - components["days"] * 8.64e+13
    
    if "hours" in allowed_units:
        components["hours"] = int(nanoseconds // 3.6e+12)
        nanoseconds = nanoseconds - components["hours"] * 3.6e+12
    
    if "minutes" in allowed_units:
        components["minutes"] = int(nanoseconds // 6e+10)
        nanoseconds = nanoseconds - components["minutes"] * 6e+10
    
    if "seconds" in allowed_units:
        components["seconds"] = int(nanoseconds // 1e+9)
        nanoseconds = nanoseconds - components["seconds"] * 1e+9
    
    if "microseconds" in allowed_units:
        components["microseconds"] = int(nanoseconds // 1_000)
        nanoseconds = nanoseconds - components["microseconds"] * 1_000
    
    # And the rest is nanoseconds
    if "nanoseconds" in allowed_units:
        components["nanoseconds"] = int(nanoseconds)

    return components

def timedelta_to_str(dt, 
                     days_in_year=360, days_in_month=30, 
                     sep=", ", mapping=None, units=None,
                     default_scope="nanoseconds"):

    
    components = timedelta_to_dict(dt, days_in_year=days_in_year, days_in_month=days_in_month, units=units)
    
    # Min and max units (components must be ordered from biggest to smallest units)
    min_unit = default_scope
    for unit, value in components.items():
        if value != 0:
            min_unit = unit
            
    max_unit = default_scope
    for unit, value in components.items():
        if value != 0:
            # Max currently found non zero
            max_unit = unit
            break
            
    
    # Determine mapping
    if mapping is None:
        mapping = {
            "years": " years",
            "months": " months",
            "weeks": " weeks",
            "days": " days",
            "hours": " hours",
            "minutes": " minutes",
            "seconds": " seconds",
            "microseconds": " microseconds",
            "nanoseconds": " nanoseconds",
        }
    elif mapping == "semishort":
        mapping = {
            "years": " years",
            "months": " months",
            "weeks": " weeks",
            "days": " days",
            "hours": " h",
            "minutes": " min",
            "seconds": " sec",
            "microseconds": " microsec",
            "nanoseconds": "nanosec",
        }
    elif mapping == "short":
        mapping = {
            "years": "Y",
            "months": "M",
            "weeks": "W",
            "days": "d",
            "hours": "h",
            "minutes": "m",
            "seconds": "s",
            "microseconds": "ms",
            "nanoseconds": "ns",
        }

    # String format
    s = ""
    is_min_reached = False
    is_max_reached = False
    for unit, value in components.items():
        is_min_reached = min_unit == unit
        is_max_reached = is_max_reached or unit == max_unit

        unit_name = mapping[unit]
        if max_unit == unit:
            s = s + f"{value}{unit_name}"
        elif is_max_reached:
            s = s + f"{sep}{value}{unit_name}"
            
        if is_min_reached:
            break
    return s

def to_timestamp(dt) -> float:
    if hasattr(dt, "to_pydatetime"):
        dt = dt.to_pydatetime()
    return dt.timestamp()


def get_period_span(period:'TimePeriod') -> Tuple[datetime.datetime, datetime.datetime]:

    # To prevent circular import
    from rocketry.parse import parse_time

    if period is None:
        return TimePeriod.min, TimePeriod.max
    elif isinstance(period, str):
        period = parse_time(period)
    dt = datetime.datetime.fromtimestamp(time.time())

    interval = period.rollback(dt)
    start = interval.left
    end = interval.right
    return start, end



import time
TOLERANCE = 2

def wait_till_task_finish(task):
    start = time.time()
    while task.status == "run":
        time.sleep(0.001)
        task.session.scheduler.handle_logs()
        if time.time() - start >= TOLERANCE:
            raise TimeoutError("Task did not finish.")
    task.session.scheduler.handle_logs()



import re
from typing import Pattern
from .base import Expression, Key

class Regex(Expression):
    """Regex expression

    Parameters
    ----------
    regex : Pattern
        Regex pattern
    key : Key
        Value of the key to match

    Examples
    --------

    .. code-block:: python

        Regex(Key('mycol'), regex=r'this .+ thing')
    """
    def __init__(self, key:Key, regex:Pattern):
        self.key = key
        self.regex = regex

    def match(self, item:dict):
        return bool(re.match(self.regex, self.key.get_value(item)))

    def __str__(self):
        return f're.match({repr(self.regex)}, {str(self.key)})'



import copy
from typing import Callable, List, Optional, Pattern, Union
from rocketry.core.parameters.parameters import Parameters
from rocketry.core.condition import BaseCondition

class FuncCond(BaseCondition):
    """Condition from a function.

    This class is to create conditions directly from
    functions.

    Parameters
    ----------
        func : callable
            Function that should return True or False 
            depending on the state of the condition.
            Can also be passed later as decorator.
        syntax : str, Patter, list
            String, regex pattern or list of such to be 
            passed to the string parser.
        args : tuple
            Argumnets to be passed to the function.
            Optional
        kwargs : dict
            Keyword arguments to be passed to the function.
            Optional

    Examples
    --------

    >>> from rocketry.conditions import FuncCond
    >>> from rocketry.parse import parse_condition

    Simple example:

    >>> @FuncCond(syntax="is foo")
    ... def is_foo():
    ...     ...
    ...     return True

    >>> parse_condition("is foo")
    FuncCond(is_foo, syntax='is foo', args=(), kwargs={})

    You can also have named parameters (using regex groups)
    to be passed to the function:

    >>> import re
    >>> @FuncCond(syntax=re.compile(r"is foo in (?P<myval>.+)"))
    ... def is_foo(myval):
    ...     ...
    ...     if myval == "house":
    ...         return True
    ...     else:
    ...         return False

    >>> parse_condition("is foo in house")
    FuncCond(is_foo, syntax=re.compile('is foo in (?P<myval>.+)'), args=(), kwargs={'myval': 'house'})
    """

    def __init__(self, 
                 func:Callable[..., bool]=None,
                 syntax:Union[str, Pattern, List[Union[str, Pattern]]]=None, 
                 args:Optional[tuple]=None, 
                 kwargs:Optional[dict]=None,
                 decor_return_func=False,
                 session=None):

        self.func = func
        self.syntax = syntax
        self.args = () if args is None else args
        self.kwargs = {} if kwargs is None else kwargs
        self.decor_return_func = decor_return_func
        if session:
            self.session = session
        if self.syntax is not None:
            self._set_parsing()

    def _recreate(self, *args, **kwargs) -> 'FuncCond':
        "Recreate the condition using args and kwargs"
        new_self = copy.copy(self)
        new_self.args = args
        new_self.kwargs = kwargs
        return new_self

    def __call__(self, func: Callable[..., bool]):
        if self.func is not None:
            raise ValueError("Function for the condition is already set")
        self.func = func
        if self.decor_return_func:
            return func # To prevent problems with pickling
        else:
            return self

    def __bool__(self):
        return self.func(*self.args, **self.kwargs)

    def observe(self, **kwargs) -> bool:
        func_params = Parameters._from_signature(self.func, **kwargs)
        param_dict = func_params.materialize(**kwargs)
        return self.get_state(*self.args, **self.kwargs, **param_dict)

    def get_state(self, *args, **kwargs):
        return self.func(*args, **kwargs)

    def _set_parsing(self):

        session = self.session

        syntaxes = [self.syntax] if not isinstance(self.syntax, (list, tuple, set)) else self.syntax
        for syntax in syntaxes:
            session._cond_parsers[syntax] = self._recreate

    def __repr__(self):
        cls_name = type(self).__name__
        func_name = self.func.__name__
        syntax = repr(self.syntax)
        return f'{cls_name}({func_name}, syntax={syntax}, args={repr(self.args)}, kwargs={repr(self.kwargs)})'



import re, time
import datetime
from typing import Tuple

from redbird.oper import in_, between

from rocketry.core.condition import All, Any
from rocketry.args import Task, Session
from rocketry.core.condition import BaseCondition
from rocketry.core.condition.base import BaseComparable
from rocketry.core.time import TimePeriod
from rocketry.core.time.utils import get_period_span, to_timestamp
from rocketry.log.utils import get_field_value

class DependMixin(BaseCondition):

    _dep_actions = None

    def __init__(self, depend_task, task=None):
        self.task = task
        self.depend_task = depend_task
        super().__init__()

    @classmethod
    def _parse_multi_all(cls, depend_tasks:str, task=None):
        from rocketry.parse import ParserError
        tasks = depend_tasks.split("', '")
        if not tasks:
            raise ParserError
        return All(*(cls(depend_task=dep_task, task=task) for dep_task in tasks))

    @classmethod
    def _parse_multi_any(cls, depend_tasks:str, task=None):
        from rocketry.parse import ParserError
        tasks = depend_tasks.split("', '")
        if not tasks:
            raise ParserError
        return Any(*(cls(depend_task=dep_task, task=task) for dep_task in tasks))

    def get_state(self, task=Task(), session=Session()):
        actual_task = session[self.task] if self.task is not None else task
        depend_task = session[self.depend_task]

        #! TODO: use Task._last_success & Task._last_run if not none and not forced
        last_depend_finish = depend_task.logger.get_latest(action=in_(self._dep_actions))
        last_actual_start = actual_task.logger.get_latest(action="run")

        if not last_depend_finish:
            # Depend has not run at all
            return False
        elif not last_actual_start:
            # Depend has succeeded but the actual task has not
            return True
            
        return get_field_value(last_depend_finish, "created") > get_field_value(last_actual_start, "created")

class TaskStatusMixin(BaseComparable):

    _action = None

    def __init__(self, period=None, task=None):
        self.task = task
        self.period = period
        super().__init__()

    def get_measurement(self, task=Task(), session=Session()):
        task = session[self.task] if self.task is not None else task
        _start_, _end_ = get_period_span(self.period if self.period is not None else task.period)
        
        allow_optimization = not self.session.config.force_status_from_logs

        if allow_optimization:
            
            # Get features that could be used to bypass reading logs
            if isinstance(self._action, str):
                last_occur = getattr(task, f'last_{self._action}')
                occurred_on_period = _start_ <= last_occur <= _end_ if last_occur is not None else False
                cannot_have_occurred = last_occur is None or last_occur < _start_
            else:
                # Multiple actions
                cannot_have_occurred = True
                for action in self._action:
                    last_occur = getattr(task, f'last_{action}')
                    if last_occur is not None and _start_ <= last_occur <= _end_:
                        occurred_on_period = True
                        cannot_have_occurred = False
                        break
                    elif last_occur is not None and last_occur >= _start_:
                        cannot_have_occurred = False
                else:
                    occurred_on_period = False
            
            # Check if can be determined without reading logs
            # NOTE: if the last_occurred > _end_, we cannot determine whether the cond is true or not
            if self._is_equal_zero():
                if cannot_have_occurred:
                    return True
                elif occurred_on_period:
                    return False
            elif self._is_any_over_zero():
                if cannot_have_occurred:
                    # If never occurred, it hasn't occurred on the period either
                    return False
                elif occurred_on_period:
                    return True

        
        records = task.logger.get_records(
            created=between(to_timestamp(_start_), to_timestamp(_end_)), 
            action=in_(self._action) if isinstance(self._action, list) else self._action
        )
        return [
            get_field_value(record, "created") 
            for record in records
        ]

    def __str__(self):
        if hasattr(self, "_str"):
            return self._str
        period = self.period
        task = self.task
        return f"task 'task '{task}' {self._action} {period}"



from rocketry.conditions import (
    TaskFinished, 
)
from rocketry.tasks import FuncTask

def run_task(fail=False):
    print("Running func")
    if fail:
        raise RuntimeError("Task failed")

def test_task_finish_compare(tmpdir, session):
    # Going to tempdir to dump the log files there
    with tmpdir.as_cwd() as old_dir:
        equals = TaskFinished(task="runned task") == 2
        greater = TaskFinished(task="runned task") > 2
        less = TaskFinished(task="runned task") < 2

        task = FuncTask(
            run_task, 
            name="runned task",
            execution="main"
        )

        # Has not yet ran
        assert not bool(equals.observe(session=session))
        assert not bool(greater.observe(session=session))
        assert bool(less.observe(session=session))

        task()
        task()

        assert bool(equals.observe(session=session))
        assert not bool(greater.observe(session=session))
        assert not bool(less.observe(session=session))

        task()
        assert not bool(equals.observe(session=session))
        assert bool(greater.observe(session=session))
        assert not bool(less.observe(session=session))



from typing import List
from collections import Counter

from .base import (
    Expression,
    QueryBase,
    Key,
    All, Any, Not,
    Equal, NotEqual, Greater, GreaterEqual, Less, LessEqual,
    true, false, Boolean
)
from .string import Regex

class Parser:
    """Parse various formats to a generic query language
    
    The query language is used to translate filtering
    from conditions to loggers and APIs to the session
    in a uniform way.
    """
    
    operators = {
        'min': GreaterEqual,
        'max': LessEqual,
        'not': NotEqual,
        'regex': Regex,
    }
    def _get_operation(self, key, val):
        if '$' in key:
            key, oper = key.split('$')
            cls_oper = self.operators[oper]
            return cls_oper(Key(key), val)
        else:
            return Equal(Key(key), val)

    def from_dict(self, d:dict):
        """Parse query from dict
        
        Examples
        --------
        >>> from rocketry.pybox.query import parser
        >>> qry = parser.from_dict({
        ...     'mydate$min': '2021-07-01',
        ...     'mydate$max': '2021-07-15',
        ...     'mykey1': 10,
        ...     'mystring$regex': r'all of .+',
        ...     'mykey2$not': 'not this',
        ... })
        >>> str(qry)
        "((<mydate> >= '2021-07-01') & (<mydate> <= '2021-07-15') & (<mykey1> == 10) & re.match('all of .+', <mystring>) & (<mykey2> != 'not this'))"
        """
        statement = None

        for key, val in d.items():
            substatement = self._get_operation(key, val)
                
            if statement is None:
                statement = substatement
            else:
                statement &= substatement

        if statement is None:
            return true
        return statement

    def from_tuples(self, l:List[tuple]):
        """Parse query from list of tuples
        
        Examples
        --------
        >>> from rocketry.pybox.query import parser
        >>> qry = parser.from_tuples([
        ...     ('mydate$min', '2021-07-01'),
        ...     ('mydate$max', '2021-07-15'),
        ...     ('mykey3', 'this'),
        ...     ('mykey3', 'that')
        ... ])
        >>> str(qry)
        "((<mydate> >= '2021-07-01') & (<mydate> <= '2021-07-15') & ((<mykey3> == 'this') | (<mykey3> == 'that')))"
        """
        key_count = Counter(item[0] for item in l)
        multi = []
        statement = None
        for key, val in l:
            substatement = self._get_operation(key, val)
            if key_count[key] > 1:
                multi.append(substatement)
                continue

            if statement is None:
                statement = substatement
            else:
                statement &= substatement

        if multi:
            substatement = Any(*multi)
            if statement is None:
                statement = substatement
            else:
                statement &= Any(*multi)

        if statement is None:
            return true
        return statement

    def from_kwargs(self, **kwargs):
        """Parse query from Pythonic keyword arguments

        Examples
        --------
        >>> from rocketry.pybox.query import parser
        >>> qry = parser.from_kwargs(
        ...     mydate=('2021-07-01', '2021-07-15'),
        ...     mykey='this'
        ... )
        >>> str(qry)
        "((<mydate> >= '2021-07-01') & (<mydate> <= '2021-07-15') & (<mykey> == 'this'))"
        """
        statement = None

        for key, val in kwargs.items():
            if isinstance(val, tuple):
                # A range
                min_, max_ = val
                if min_ is not None and max_ is not None:
                    substatement = GreaterEqual(Key(key), min_) & LessEqual(Key(key), max_)
                elif min_ is not None:
                    substatement = GreaterEqual(Key(key), min_)
                elif max_ is not None:
                    substatement = LessEqual(Key(key), max_)
                else:
                    continue
            elif isinstance(val, list):
                substatement = Any(*(Key(key) == subval for subval in val))
            else:
                substatement = Equal(Key(key), val)
                
            if statement is None:
                statement = substatement
            else:
                statement &= substatement

        if statement is None:
            return true
        return statement




import datetime

from rocketry.conditions import (
    SchedulerStarted,
    SchedulerCycles
)
from rocketry.time import (
    TimeDelta, 
    TimeOfDay
)

def test_scheduler_cycles(session):

    session.config.shut_cond = SchedulerCycles.from_magic(__eq__=3)
    session.start()
    # Imitating the __bool__
    assert (SchedulerCycles() == 3).observe(session=session)
    assert (SchedulerCycles() > 2).observe(session=session)
    assert (SchedulerCycles() < 4).observe(session=session)
    assert (SchedulerCycles() != 4).observe(session=session)
    assert not (SchedulerCycles() < 3).observe(session=session)
    assert not (SchedulerCycles() > 3).observe(session=session)
    assert not (SchedulerCycles() == 4).observe(session=session)


def test_scheduler_started(session):

    session.scheduler.startup_time = datetime.datetime.now() - datetime.timedelta(0, 20, 0) # 20 seconds ago
    # Imitating the __bool__
    assert (SchedulerStarted(period=TimeDelta("30 seconds"))).observe(session=session)
    assert not (SchedulerStarted(period=TimeDelta("10 seconds"))).observe(session=session)




import calendar
import datetime
import re

import dateutil

from rocketry.core.time.anchor import AnchoredInterval
from rocketry.core.time.base import TimeInterval
from rocketry.core.time.utils import timedelta_to_str, to_dict, to_nanoseconds
from rocketry.pybox.time.interval import Interval


class TimeOfMinute(AnchoredInterval):
    """Time interval anchored to minute cycle of a clock

    min: 0 seconds, 0 microsecond, 0 nanosecond
    max: 59 seconds, 999999 microsecond, 999 nanosecond

    Example:
        # From 5th second of a minute to 30th second of a minute
        TimeOfHour("5:00", "30:00")
    """

    _scope = "minute"
    _scope_max = to_nanoseconds(minute=1) - 1
    _unit_resolution = to_nanoseconds(second=1)

    def anchor_str(self, s, **kwargs):
        # ie. 30.123
        res = re.search(r"(?P<second>[0-9][0-9])([.](?P<microsecond>[0-9]{0,6}))?(?P<nanosecond>[0-9]+)?", s, flags=re.IGNORECASE)
        if res:
            if res["microsecond"] is not None:
                res["microsecond"] = res["microsecond"].ljust(6, "0")
            return to_nanoseconds(**{key: int(val) for key, val in res.groupdict().items() if val is not None})

        res = re.search(r"(?P<n>[1-4] ?(quarter|q))", s, flags=re.IGNORECASE)
        if res:
            # ie. "1 quarter"
            n_quarters = res["n"]
            return (self._scope_max + 1) / 4 * n_quarters - 1


class TimeOfHour(AnchoredInterval):
    """Time interval anchored to hour cycle of a clock

    min: 0 minutes, 0 seconds, 0 microsecond, 0 nanosecond
    max: 59 minutes, 59 seconds, 999999 microsecond, 999 nanosecond

    Example:
        # From 15 past to half past
        TimeOfHour("15:00", "30:00")
    """
    _scope = "hour"
    _scope_max = to_nanoseconds(hour=1) - 1
    _unit_resolution = to_nanoseconds(minute=1)

    def anchor_str(self, s, **kwargs):
        # ie. 12:30.123
        res = re.search(r"(?P<minute>[0-9][0-9]):(?P<second>[0-9][0-9])([.](?P<microsecond>[0-9]{0,6}))?(?P<nanosecond>[0-9]+)?", s, flags=re.IGNORECASE)
        if res:
            if res["microsecond"] is not None:
                res["microsecond"] = res["microsecond"].ljust(6, "0")
            return to_nanoseconds(**{key: int(val) for key, val in res.groupdict().items() if val is not None})

        res = re.search(r"(?P<n>[1-4]) ?(quarter|q)", s, flags=re.IGNORECASE)
        if res:
            # ie. "1 quarter"
            n_quarters = int(res["n"])
            return (self._scope_max + 1) / 4 * n_quarters - 1


class TimeOfDay(AnchoredInterval):
    """Time interval anchored to day cycle of a clock
    
    min: 0 hours, 0 minutes, 0 seconds, 0 microsecond, 0 nanosecond
    max: 23 hours, 59 minutes, 59 seconds, 999999 microsecond, 999 nanosecond

    Example:
        # From 10 o'clock to 15 o'clock
        TimeOfDay("10:00", "15:00")
    """
    _scope = "day"
    _scope_max = to_nanoseconds(day=1) - 1
    _unit_resolution = to_nanoseconds(hour=1)

    def anchor_str(self, s, **kwargs):
        # ie. "10:00:15"
        dt = dateutil.parser.parse(s)
        d = to_dict(dt)
        components = ("hour", "minute", "second", "microsecond", "nanosecond")
        return to_nanoseconds(**{key: int(val) for key, val in d.items() if key in components})

    def anchor_dt(self, dt, **kwargs):
        "Turn datetime to nanoseconds according to the scope (by removing higher time elements)"
        d = to_dict(dt)
        d = {
            key: val
            for key, val in d.items()
            if key in ("hour", "minute", "second", "microsecond", "nanosecond")
        }
        return to_nanoseconds(**d) 

class TimeOfWeek(AnchoredInterval):
    """Time interval anchored to week cycle
    
    min: Monday, 0 hours, 0 minutes, 0 seconds, 0 microsecond, 0 nanosecond
    max: Sunday, 23 hours, 59 minutes, 59 seconds, 999999 microsecond, 999 nanosecond

    Example:
        # From Monday 3 PM to Wednesday 4 PM
        TimeOfWeek("Mon 15:00", "Wed 16:00")
    """
    _scope = "week"
    _scope_max = to_nanoseconds(day=7) - 1 # Sun day end of day 
    _unit_resolution = to_nanoseconds(day=1)

    weeknum_mapping = {
        **dict(zip(calendar.day_name, range(7))), 
        **dict(zip(calendar.day_abbr, range(7))), 
        **dict(zip(range(7), range(7)))
    }
    # TODO: ceil end
    def anchor_str(self, s, side=None, **kwargs):
        # Allowed:
        #   "Mon", "Monday", "Mon 10:00:00"
        res = re.search(r"(?P<dayofweek>[a-z]+) ?(?P<time>.*)", s, flags=re.IGNORECASE)
        comps = res.groupdict()
        dayofweek = comps.pop("dayofweek")
        time = comps.pop("time")
        nth_day = self.weeknum_mapping[dayofweek]

        # TODO: TimeOfDay.anchor_str as function
        if not time:
            nanoseconds = to_nanoseconds(day=1) - 1 if side == "end" else 0
        else:
            nanoseconds = TimeOfDay().anchor_str(time) 

        return to_nanoseconds(day=1) * nth_day + nanoseconds

    def anchor_dt(self, dt, **kwargs):
        "Turn datetime to nanoseconds according to the scope (by removing higher time elements)"
        d = to_dict(dt)
        d = {
            key: val
            for key, val in d.items()
            if key in ("hour", "minute", "second", "microsecond", "nanosecond")
        }
        dayofweek = dt.weekday()
        return to_nanoseconds(**d) + dayofweek * to_nanoseconds(day=1)


class TimeOfMonth(AnchoredInterval):
    """Time interval anchored to day cycle of a clock
    
    min: first day of month, 0 hours, 0 minutes, 0 seconds, 0 microsecond, 0 nanosecond
    max: last day of month, 23 hours, 59 minutes, 59 seconds, 999999 microsecond, 999 nanosecond

    Example:
        # From 10 o'clock to 15 o'clock
        TimeOfMonth("1st", "5th")
    """
    # TODO: Support for month end (ie. last 5th day of month to last 2nd)
    # Could be implemented by allowing minus _start and minus _end
    #   rollforward/rollback/contains would need slight changes 

    _scope = "year"
    _scope_max = to_nanoseconds(day=31) - 1 # 31st end of day
    _unit_resolution = to_nanoseconds(day=1)
     # NOTE: Floating
    # TODO: ceil end and implement reversion (last 5th day)

    def anchor_str(self, s, side=None, **kwargs):
        # Allowed:
        #   "5th", "1st", "5", "3rd 12:30:00"
        #   TODO: "first 5th", "last 5th", "last 3rd 10:00:00"
        res = re.search(r"(?P<dayofmonth>[0-9]+)(st|nd|rd|th|[.])? ?(?P<time>.*)", s, flags=re.IGNORECASE)
        comps = res.groupdict()
        dayofmonth = comps.pop("dayofmonth")
        time = comps.pop("time")
        nth_day = int(dayofmonth)

        # TODO: TimeOfDay.anchor_str as function
        ceil_time = not time and side == "end"
        if ceil_time:
            # If time is not defined and the end
            # is being anchored, the time is ceiled.

            # If one says 'thing X was organized between 
            # 15th and 17th of July', the sentence
            # includes 17th till midnight.
            nanoseconds = to_nanoseconds(day=1) - 1
        elif time:
            nanoseconds = TimeOfDay().anchor_str(time) 
        else:
            nanoseconds = 0

        return to_nanoseconds(day=1) * (nth_day - 1) + nanoseconds

    def anchor_int(self, i, **kwargs):
        return i * self._unit_resolution

    def anchor_dt(self, dt, **kwargs):
        "Turn datetime to nanoseconds according to the scope (by removing higher time elements)"
        d = to_dict(dt)
        d = {
            key: val
            for key, val in d.items()
            if key in ("day", "hour", "minute", "second", "microsecond", "nanosecond")
        }
        if "day" in d:
            # Day (of month) does not start from 0 (but from 1)
            d["day"] = d["day"] - 1

        return to_nanoseconds(**d)

    def get_scope_forward(self, dt):
        n_days = calendar.monthrange(dt.year, dt.month)[1]
        return to_nanoseconds(day=1) * n_days

    def get_scope_back(self, dt):
        month = 12 if dt.month == 1 else dt.month - 1
        year = dt.year - 1 if dt.month == 1 else dt.year
        n_days = calendar.monthrange(year, month)[1]
        return to_nanoseconds(day=1) * n_days

class TimeOfYear(AnchoredInterval):
    """Time interval anchored to day cycle of a clock

    min: 1st Jan, 0 hours, 0 minutes, 0 seconds, 0 microsecond, 0 nanosecond
    max: 31st Dec, 23 hours, 59 minutes, 59 seconds, 999999 microsecond, 999 nanosecond

    Example:
        # From 10 o'clock to 15 o'clock
        TimeOfYear("Jan", "Feb")
    """

    _scope = "year"
    _scope_max = to_nanoseconds(day=1) * 366 - 1

    monthnum_mapping = {
        **dict(zip(calendar.month_name[1:], range(12))), 
        **dict(zip(calendar.month_abbr[1:], range(12))), 
        **dict(zip(range(12), range(12)))
    }
    # NOTE: Floating

    def anchor_str(self, s, side=None, **kwargs):
        # Allowed:
        #   "January", "Dec", "12", "Dec last 5th 10:00:00"
        res = re.search(r"(?P<monthofyear>[a-z]+) ?(?P<month>.*)", s, flags=re.IGNORECASE)
        comps = res.groupdict()
        monthofyear = comps.pop("monthofyear")
        month_str = comps.pop("month")
        nth_month = self.monthnum_mapping[monthofyear]

        # TODO: TimeOfDay.anchor_str as function
        # nanoseconds = TimeOfMonth().anchor_str(month_str) if month_str else 0

        ceil_time = not month_str and side == "end"
        if ceil_time:
            # If time is not defined and the end
            # is being anchored, the time is ceiled.

            # If one says 'thing X was organized between 
            # 15th and 17th of July', the sentence
            # includes 17th till midnight.
            nanoseconds = to_nanoseconds(day=31) - 1 
        elif month_str:
            nanoseconds = TimeOfMonth().anchor_str(month_str) 
        else:
            nanoseconds = 0

        return nth_month * to_nanoseconds(day=31) + nanoseconds

    def anchor_dt(self, dt, **kwargs):
        "Turn datetime to nanoseconds according to the scope (by removing higher time elements)"
        dt_dict = to_dict(dt)
        d = {
            key: val
            for key, val in dt_dict.items()
            if key in ("day", "hour", "minute", "second", "microsecond", "nanosecond")
        }
        nth_month = dt_dict["month"] - 1
        if "day" in d:
            # Day (of month) does not start from 0 (but from 1)
            d["day"] = d["day"] - 1
        return nth_month * to_nanoseconds(day=31) + to_nanoseconds(**d)


class RelativeDay(TimeInterval):
    """Specific day

    Examples:
    ---------
        Day("today")
        Day("yesterday")
        Day("yesterday")
    """

    offsets = {
        "today": datetime.timedelta(),
        "yesterday": datetime.timedelta(days=1),
        "the_day_before":datetime.timedelta(days=2),
        #"first_day_of_year": get_first_day_of_year,
    }

    min_time = datetime.time.min
    max_time = datetime.time.max

    def __init__(self, day, *, start_time=None, end_time=None):
        self.day = day
        self.start_time = self.min_time if start_time is None else start_time
        self.end_time = self.max_time if end_time is None else end_time

    def rollback(self, dt):
        offset = self.offsets[self.day]
        dt = dt - offset
        return Interval(
            datetime.datetime.combine(dt, self.start_time),
            datetime.datetime.combine(dt, self.end_time)
        )

    def rollforward(self, dt):
        raise AttributeError("RelativeDay has no next day")

    def __repr__(self):
        args_str = str(self.day)
        if self.start_time != self.min.date():
            args_str += f', start_time={self.start_time}'
        if self.end_time != self.max.date():
            args_str += f', end_time={self.end_time}'

        return f"RelativeDay({args_str})"



class SchedulerRestart(Exception):
    "Restart the scheduler"


class SchedulerExit(Exception):
    "Shut down the scheduler"


class TaskInactionException(Exception):
    "Task did not succeed nor fail. It did not need to run."


class TaskTerminationException(Exception):
    """Task was terminated.
    This should only be raised by threaded
    tasks to signal that they did indeed
    listen the thread_please_terminate event
    and ended as a result of that"""



import re

from rocketry.core import BaseCondition
from rocketry.conditions import FuncCond
from rocketry.parse.condition import parse_condition

def test_func_cond():

    @FuncCond(syntax="is foo")
    def is_foo():
        return True
    
    cond = parse_condition("is foo")
    assert isinstance(cond, BaseCondition)
    assert bool(cond)

def test_func_cond_with_kwargs():

    @FuncCond(syntax=re.compile("is foo (?P<myval>true|false)"))
    def is_foo(myval):
        return True if myval == "true" else False if myval == "false" else None
    
    cond_true = parse_condition("is foo true")
    assert isinstance(cond_true, BaseCondition)
    assert bool(cond_true)

    cond_false = parse_condition("is foo false")
    assert isinstance(cond_false, BaseCondition)
    assert not bool(cond_false)
    
    assert cond_true is not cond_false




import logging
from rocketry.conditions.scheduler import SchedulerStarted
from rocketry.core import Scheduler
from rocketry.conditions import TaskStarted, AlwaysTrue

from redbird.logging import RepoHandler
from redbird.repos import MemoryRepo
from rocketry.core.time.base import TimeDelta

from rocketry.log.log_record import LogRecord
from rocketry.tasks import FuncTask

import pytest


@pytest.mark.parametrize("execution", ["main", "thread", "process"])
@pytest.mark.parametrize(
    "script_path,expected_outcome,exc_cls",
    [
        pytest.param(
            "scripts/succeeding_script.py", 
            "success",
            None,
            id="Success"),
        pytest.param(
            "scripts/failing_script.py", 
            "fail", 
            RuntimeError,
            id="Failure"),
    ],
)
def test_run(tmpdir, script_files, script_path, expected_outcome, exc_cls, execution, session):
    with tmpdir.as_cwd() as old_dir:
        task_logger = logging.getLogger(session.config.task_logger_basename)
        task_logger.handlers = [
            RepoHandler(repo=MemoryRepo(model=LogRecord))
        ]
        task = FuncTask(
            func_name="main",
            path=script_path, 
            name="a task",
            start_cond=AlwaysTrue(),
            execution=execution
        )
        
        session.config.shut_cond = (TaskStarted(task="a task") >= 3) | ~SchedulerStarted(period=TimeDelta("15 seconds"))
        session.start()

        if expected_outcome == "fail":
            failures = list(task.logger.filter_by(action="fail").all())
            assert 3 == len(failures)

            # Check it has correct traceback in message
            for fail in failures:
                tb = fail.exc_text
                assert "Traceback (most recent call last):" in tb
                assert "RuntimeError: This task failed" in tb
        else:
            success = list(task.logger.filter_by(action="success").all())
            assert 3 == len(success)





from typing import Type

def _add_parser(cls:Type, container:dict) -> None:
    """Acquire the parsers from the class
    and add them to the container.
    """
    attr_parsers = "__parsers__"
    def _get_constructor(val, cls):
        if not isinstance(val, str):
            # If not string, e xpecting a function/callable
            return val
        if val == "__init__":
            # cls.__init__ would require the instance, self so we use basic constructor
            return cls
        else:
            # Expecting a class method
            return getattr(cls, val)

    parsers = getattr(cls, attr_parsers, None)
    if not parsers:
        return

    parsers = {
        # Typically a callable is expected but if string given, a method is assumed
        key: _get_constructor(val, cls=cls)
        for key, val in parsers.items()
    }

    container.update(parsers)
    # We delete the attribute to not cause confusion 
    # as changing it does nothing
    delattr(cls, attr_parsers)

def _register(cls, container:dict):
    """Add class to the container"""
    attr_register = "__register__"

    if getattr(cls, attr_register, True):
        container[cls.__name__] = cls
    if hasattr(cls, attr_register):
        delattr(cls, attr_register)


import platform
import pytest

from rocketry.args import Private, Return
from rocketry import Scheduler
from rocketry.conditions.scheduler import SchedulerCycles
from rocketry.core import parameters
from rocketry.tasks import FuncTask
from rocketry.conditions import TaskStarted, AlwaysTrue
from rocketry.args import FuncArg, Arg


def get_x():
    return "x"

def get_y():
    return "y"

def func_x_with_arg(myparam):
    assert myparam == "x"

def get_with_nested_args(arg = FuncArg(get_y), arg_2 = Arg('session_arg')):
    assert arg == "y"
    assert arg_2 == "z"
    return 'x'

@pytest.mark.parametrize("execution", ["main", "thread", "process"])
def test_simple(session, execution):

    task = FuncTask(
        func_x_with_arg, 
        parameters={"myparam": FuncArg(get_x)}, 
        execution=execution, 
        name="a task", 
        start_cond=AlwaysTrue()
    )
    session.config.shut_cond = (TaskStarted(task="a task") >= 1)

    assert task.status is None
    session.start()

    assert "success" == task.status

@pytest.mark.parametrize("execution", ["main", "thread", "process"])
def test_session(session, execution):

    session.parameters["myparam"] = FuncArg(get_x)

    task = FuncTask(
        func_x_with_arg, 
        execution=execution, 
        name="a task", 
        start_cond=AlwaysTrue()
    )
    session.config.shut_cond = (TaskStarted(task="a task") >= 1)

    assert task.status is None
    session.start()

    assert "success" == task.status

@pytest.mark.parametrize("execution", ["main", "thread", "process"])
def test_session_with_arg(session, execution):

    session.parameters["a_param"] = FuncArg(get_x)

    task = FuncTask(
        func_x_with_arg, 
        execution=execution,
        name="a task", 
        parameters={"myparam": Arg('a_param')},
        start_cond=AlwaysTrue()
    )
    session.config.shut_cond = (TaskStarted(task="a task") >= 1)

    assert task.status is None
    session.start()
    assert "success" == task.status

@pytest.mark.parametrize("config_mater", ['pre', 'post', None])
@pytest.mark.parametrize("materialize", ['pre', 'post', None])
@pytest.mark.parametrize("execution", ["main", "thread", "process"])
def test_nested(session, execution, materialize, config_mater):
    if config_mater is not None:
        session.config.param_materialize = config_mater

    session.parameters["session_arg"] = "z"
    session.parameters["myparam"] = FuncArg(get_with_nested_args, materialize=materialize)

    task = FuncTask(
        func_x_with_arg, 
        execution=execution, 
        name="a task", 
        start_cond=AlwaysTrue()
    )
    session.config.shut_cond = (TaskStarted(task="a task") >= 1)

    assert task.status is None
    session.start()
    if platform.system() == "Windows" and execution == "process" and (materialize == "post" or (materialize is None and config_mater in ("post", None))):
        # Windows cannot pickle the session but apparently Linux can
        assert "fail" == task.status
    else:
        assert "success" == task.status

class UnPicklable:
    def __getstate__(self):
        raise RuntimeError("Cannot be pickled")

def get_unpicklable():
    return UnPicklable()

def func_check_unpicklable(myparam):
    assert isinstance(myparam, UnPicklable)

@pytest.mark.parametrize("execution", ["process"])
def test_unpicklable(session, execution):
    "Test a FuncArg that returns unpicklable item"
    task = FuncTask(
        func_check_unpicklable, 
        parameters={"myparam": FuncArg(get_unpicklable)}, 
        execution=execution, 
        name="a task", 
        start_cond=AlwaysTrue()
    )

    session.config.shut_cond = (TaskStarted(task="a task") >= 1) 

    assert task.status is None
    session.start()
    assert "success" == task.status



from rocketry.parse import parse_condition
from rocketry.conditions import (
    AlwaysTrue, AlwaysFalse, 
    Not,
)
import pytest

@pytest.mark.parametrize(
    "cond_str,expected",
    [
        pytest.param(
            "always true", 
            AlwaysTrue(),
            id="AlwaysTrue"),

        pytest.param(
            "always false", 
            AlwaysFalse(),
            id="AlwaysFalse"),

        pytest.param(
            "always true & always true", 
            AlwaysTrue() & AlwaysTrue(),
            id="All"),
        pytest.param(
            "always true | always true", 
            AlwaysTrue() | AlwaysTrue(),
            id="Any"),
        pytest.param(
            "~always true", 
            Not(AlwaysTrue()),
            id="Not"),

        pytest.param(
            "~~always true", 
            ~~AlwaysTrue(),
            id="Nested Not"),

        pytest.param(
            "~(always true)", 
            ~AlwaysTrue(),
            id="Closured Not"),

        pytest.param(
            "~(always true & always true)", 
            ~(AlwaysTrue() & AlwaysTrue()),
            id="All in Not"),

        pytest.param(
            "~(always true | always true)", 
            ~(AlwaysTrue() | AlwaysTrue()),
            id="Any in Not"),

        pytest.param(
            "~always true | ~always true & ~always true", 
            ~AlwaysTrue() | ~AlwaysTrue() & ~AlwaysTrue(),
            id="Multiple Not"),

        pytest.param(
            "always true & (always true | always true & (always true | always true))", 
            AlwaysTrue() & (AlwaysTrue() | AlwaysTrue() & (AlwaysTrue() | AlwaysTrue())),
            id="Deeply nested"),
        pytest.param(
            "~always true & ~~(always true | ~always true & ~(~~always true | ~always true))", 
            ~AlwaysTrue() & ~~(AlwaysTrue() | ~AlwaysTrue() & ~(~~AlwaysTrue() | ~AlwaysTrue())),
            id="Deeply nested Not"),

        pytest.param(
            "", 
            AlwaysFalse(),
            id="Empty", marks=pytest.mark.xfail),
    ],
)
def test_string(cond_str, expected):
    cond = parse_condition(cond_str)
    assert cond == expected


@pytest.mark.parametrize(
    "cond_str",
    [
        pytest.param(
            "daily between 20:00 and 23:59", 
            id="daily between"),
    ],
)
def test_string_matching(cond_str):
    cond = parse_condition(cond_str)



import asyncio
import datetime
import time
import os

import pytest
from rocketry.conditions.scheduler import SchedulerStarted
from rocketry.conditions.task import TaskTerminated

from rocketry.core import Scheduler
from rocketry.core.time.base import TimeDelta
from rocketry.tasks import FuncTask
from rocketry.exc import TaskTerminationException
from rocketry.conditions import TaskFinished, TaskStarted, AlwaysTrue, AlwaysFalse
from rocketry.args import Session

def run_slow():
    time.sleep(1)
    with open("work.txt", "a") as file:
        file.write("line created\n")

def run_slow_threaded(_thread_terminate_):
    time.sleep(0.2)
    if _thread_terminate_.is_set():
        raise TaskTerminationException
    else:
        with open("work.txt", "a") as file:
            file.write("line created\n")

async def run_slow_async():
    await asyncio.sleep(0.2)
    with open("work.txt", "a") as file:
        file.write("line created\n")

def get_slow_func(execution):
    return {
        "async": run_slow_async,
        "process": run_slow,
        # Thread tasks are terminated inside the task (the task should respect _thread_terminate_)
        "thread": run_slow_threaded,
    }[execution]

@pytest.mark.parametrize("execution", ["async", "thread", "process"])
def test_without_timeout(tmpdir, execution, session):
    """Test the task.timeout is respected overt scheduler.timeout"""
    # TODO: There is probably better ways to test this
    with tmpdir.as_cwd() as old_dir:

        func_run_slow = get_slow_func(execution)
        task = FuncTask(func_run_slow, name="slow task but passing", start_cond=AlwaysTrue(), timeout="never", execution=execution)

        session.config.shut_cond = (TaskFinished(task="slow task but passing") >= 2) | ~SchedulerStarted(period=TimeDelta("5 seconds"))
        session.config.timeout = 0.1
        session.start()

        logger = task.logger
        # If Scheduler is quick, it may launch the task 3 times 
        # but there still should not be any terminations
        assert 2 <= logger.filter_by(action="run").count() 
        assert 0 == logger.filter_by(action="terminate").count()
        assert 2 <= logger.filter_by(action="success").count()
        assert 0 == logger.filter_by(action="fail").count()

        assert os.path.exists("work.txt")

@pytest.mark.parametrize("execution", ["async", "thread", "process"])
def test_task_timeout_set_in_session(tmpdir, execution, session):
    """Test task termination due to the task ran too long"""
    with tmpdir.as_cwd() as old_dir:

        func_run_slow = get_slow_func(execution)

        task = FuncTask(func_run_slow, name="slow task", start_cond=AlwaysTrue(), execution=execution)

        session.config.shut_cond = (TaskStarted(task="slow task") >= 2) | ~SchedulerStarted(period=TimeDelta("5 seconds"))
        session.config.timeout = 0.1
        assert session.config.timeout == datetime.timedelta(milliseconds=100)
        session.start()

        logger = task.logger
        assert 2 == logger.filter_by(action="run").count() 
        assert 2 == logger.filter_by(action="terminate").count()
        assert 0 == logger.filter_by(action="success").count()
        assert 0 == logger.filter_by(action="fail").count()

        assert not os.path.exists("work.txt")

@pytest.mark.parametrize("execution", ["async", "thread", "process"])
def test_task_timeout_set_in_task(tmpdir, execution, session):
    """Test task termination due to the task ran too long"""
    with tmpdir.as_cwd() as old_dir:

        func_run_slow = get_slow_func(execution)

        task = FuncTask(func_run_slow, name="slow task", start_cond=AlwaysTrue(), timeout="0.1 sec", execution=execution)

        session.config.shut_cond = (TaskStarted(task="slow task") >= 2) | ~SchedulerStarted(period=TimeDelta("5 seconds"))
        assert task.timeout == datetime.timedelta(milliseconds=100)
        session.start()

        logger = task.logger
        assert 2 == logger.filter_by(action="run").count() 
        assert 2 == logger.filter_by(action="terminate").count()
        assert 0 == logger.filter_by(action="success").count()
        assert 0 == logger.filter_by(action="fail").count()

        assert not os.path.exists("work.txt")

@pytest.mark.parametrize("execution", ["async", "thread", "process"])
def test_task_terminate(tmpdir, execution, session):
    """Test task termination due to the task was terminated by another task"""

    def terminate_task(session=Session()):
        session["slow task"].force_termination = True

    with tmpdir.as_cwd() as old_dir:

        func_run_slow = get_slow_func(execution)
        task = FuncTask(func_run_slow, name="slow task", start_cond=AlwaysTrue(), execution=execution)

        FuncTask(terminate_task, name="terminator", start_cond=TaskStarted(task="slow task"), execution="main")
        session.config.shut_cond = (TaskStarted(task="slow task") >= 2) | ~SchedulerStarted(period=TimeDelta("5 seconds"))
        session.start()

        logger = task.logger
        assert 2 == logger.filter_by(action="run").count() 
        assert 2 == logger.filter_by(action="terminate").count()
        assert 0 == logger.filter_by(action="success").count()
        assert 0 == logger.filter_by(action="fail").count()

        assert not os.path.exists("work.txt")

        # Attr force_termination should be reseted every time the task has been terminated
        assert not task.force_termination


@pytest.mark.parametrize("execution", ["async", "thread", "process"])
def test_task_terminate_end_cond(tmpdir, execution, session):
    """Test task termination due to the task ran too long"""
    #! NOTE: CI observed to get stuck in this for some times
    with tmpdir.as_cwd() as old_dir:

        func_run_slow = get_slow_func(execution)

        task = FuncTask(func_run_slow, name="slow task", start_cond=AlwaysTrue(), end_cond=TaskStarted(task='slow task'), execution=execution)

        session.config.shut_cond = (TaskTerminated(task="slow task") >= 1) | ~SchedulerStarted(period=TimeDelta("5 seconds"))
        session.start()

        logger = task.logger
        assert 1 <= logger.filter_by(action="run").count() 
        assert 1 <=logger.filter_by(action="terminate").count()
        assert 0 == logger.filter_by(action="success").count()
        assert 0 == logger.filter_by(action="fail").count()

        assert not os.path.exists("work.txt")



import pytest
import datetime

from rocketry.core.time.base import (
    All, Any
)
from rocketry.time.interval import TimeOfDay

from_iso = datetime.datetime.fromisoformat

@pytest.mark.parametrize(
    "dt,periods,roll_start,roll_end",
    [
        # Regular
        pytest.param(
            from_iso("2020-01-01 07:00:00"),
            [
                TimeOfDay("08:00", "18:00"),
                TimeOfDay("10:00", "14:00"),
                TimeOfDay("12:00", "16:00"),
            ],
            from_iso("2020-01-01 12:00:00"), from_iso("2020-01-01 14:00:00"),
            id="Left from interval"),
    ],
)
def test_rollforward_all(dt, periods, roll_start, roll_end):
    time = All(*periods)

    interval = time.rollforward(dt)
    assert roll_start == interval.left
    assert roll_end == interval.right

@pytest.mark.parametrize(
    "dt,periods,roll_start,roll_end",
    [
        # Regular
        pytest.param(
            from_iso("2020-01-02 07:00:00"),
            [
                TimeOfDay("08:00", "18:00"),
                TimeOfDay("10:00", "14:00"),
                TimeOfDay("12:00", "16:00"),
            ],
            from_iso("2020-01-01 12:00:00"), from_iso("2020-01-01 14:00:00"),
            id="Left from interval"),
    ],
)
def test_rollback_all(dt, periods, roll_start, roll_end):
    time = All(*periods)

    interval = time.rollback(dt)
    assert roll_start == interval.left
    assert roll_end == interval.right

@pytest.mark.parametrize(
    "dt,periods,roll_start,roll_end",
    [
        # Regular
        pytest.param(
            from_iso("2020-01-01 07:00:00"),
            [
                TimeOfDay("08:00", "18:00"),
                TimeOfDay("10:00", "14:00"),
                TimeOfDay("12:00", "16:00"),
            ],
            from_iso("2020-01-01 08:00:00"), from_iso("2020-01-01 18:00:00"),
            id="Left from interval"),
    ],
)
def test_rollforward_any(dt, periods, roll_start, roll_end):
    time = Any(*periods)

    interval = time.rollforward(dt)
    assert roll_start == interval.left
    assert roll_end == interval.right

@pytest.mark.parametrize(
    "dt,periods,roll_start,roll_end",
    [
        # Regular
        pytest.param(
            from_iso("2020-01-02 07:00:00"),
            [
                TimeOfDay("08:00", "18:00"),
                TimeOfDay("10:00", "14:00"),
                TimeOfDay("12:00", "16:00"),
            ],
            from_iso("2020-01-01 08:00:00"), from_iso("2020-01-01 18:00:00"),
            id="Left from interval"),
    ],
)
def test_rollback_any(dt, periods, roll_start, roll_end):
    time = Any(*periods)

    interval = time.rollback(dt)
    assert roll_start == interval.left
    assert roll_end == interval.right



import multiprocessing

from rocketry.core import Scheduler
from rocketry.tasks import FuncTask
from rocketry.time import TimeDelta
from rocketry.conditions import SchedulerStarted, TaskStarted, AlwaysTrue

def run_succeeding():
    pass

def run_creating_child():

    proc = multiprocessing.Process(target=run_succeeding, daemon=True)
    proc.start()

def test_creating_child(tmpdir, session):
    with tmpdir.as_cwd() as old_dir:

        # To be confident the scheduler won't lie to us
        # we test the task execution with a job that has
        # actual measurable impact outside rocketry
        FuncTask(run_creating_child, name="task_1", start_cond=AlwaysTrue())
 
        session.config.tasks_as_daemon = False
        session.config.shut_cond = (TaskStarted(task="task_1") >= 1) | ~SchedulerStarted(period=TimeDelta("1 second"))

    
        session.start()

        logger = session.get_task("task_1").logger
        assert 1 == logger.filter_by(action="run").count()
        assert 1 == logger.filter_by(action="success").count()
        assert 0 == logger.filter_by(action="fail").count()



from rocketry.tasks import FuncTask
from rocketry.core import Parameters, Scheduler

def test_run(tmpdir, session):
    task1 = FuncTask(
        lambda : None, 
        name="example 1",
        execution="main"
    )
    
    task2 = FuncTask(
        lambda : None, 
        name="example 2",
        execution="main"
    )
    session.run("example 2")
    assert task1.status is None
    assert task2.status == "success"

def test_run_obey_cond(tmpdir, session):
    task1 = FuncTask(
        lambda : None, 
        name="example 1",
        execution="main"
    )
    
    task2 = FuncTask(
        lambda : None, 
        name="example 2",
        execution="main"
    )
    session.run("example 2", obey_cond=True)
    assert task1.status is None
    assert task2.status is None # start cond is false



import pytest
from rocketry.time.interval import (
    TimeOfDay,
    TimeOfMonth,
    TimeOfWeek
)

NS_IN_SECOND = 1e+9
NS_IN_MINUTE = 1e+9 * 60
NS_IN_HOUR   = 1e+9 * 60 * 60
NS_IN_DAY    = 1e+9 * 60 * 60 * 24

def pytest_generate_tests(metafunc):
    if metafunc.cls is not None:

        method_name = metafunc.function.__name__
        cls = metafunc.cls
        params = []
        if method_name == "test_closed":
            params = cls.scen_closed
        elif method_name == "test_open":
            return
        elif method_name == "test_open_left":
            params = cls.scen_open_left
        elif method_name == "test_open_right":
            params = cls.scen_open_right
        elif method_name == "test_time_point":
            params = cls.scen_time_point
        else:
            return
        
        argnames = [arg for arg in metafunc.fixturenames if arg in ('start', 'end', "expected_start", "expected_end", "time_point")]

        idlist = []
        argvalues = []
        argvalues = []
        for scen in params:
            idlist.append(scen.pop("id", f'{scen.get("start")}, {scen.get("end")}'))
            argvalues.append(tuple(scen[name] for name in argnames))
        metafunc.parametrize(argnames, argvalues, ids=idlist, scope="class")

class ConstructTester:

    def test_closed(self, start, end, expected_start, expected_end):
        time = self.cls(start, end)
        assert expected_start == time._start
        assert expected_end == time._end

    def test_open(self):
        time = self.cls(None, None)
        assert 0 == time._start
        assert self.max_ns == time._end

    def test_open_left(self, end, expected_end, **kwargs):
        time = self.cls(None, end)
        assert 0 == time._start
        assert expected_end == time._end

    def test_open_right(self, start, expected_start, **kwargs):
        time = self.cls(start, None)
        assert expected_start == time._start
        assert self.max_ns == time._end

    def test_time_point(self, start, expected_start, expected_end, **kwargs):
        time = self.cls(start, time_point=True)
        assert expected_start == time._start
        assert expected_end == time._end

class TestTimeOfDay(ConstructTester):

    cls = TimeOfDay

    max_ns = 24 * NS_IN_HOUR - 1

    scen_closed = [
        {
            "start": "10:00",
            "end": "12:00",
            "expected_start": 10 * NS_IN_HOUR,
            "expected_end": 12 * NS_IN_HOUR,
        }
    ]

    scen_open_left = [
        {
            "end": "12:00",
            "expected_end": 12 * NS_IN_HOUR
        }
    ]
    scen_open_right = [
        {
            "start": "12:00",
            "expected_start": 12 * NS_IN_HOUR
        }
    ]
    scen_time_point = [
        {
            "start": "12:00",
            "expected_start": 12 * NS_IN_HOUR,
            "expected_end": 13 * NS_IN_HOUR - 1,
        }
    ]


class TestTimeOfWeek(ConstructTester):

    cls = TimeOfWeek

    max_ns = 7 * NS_IN_DAY - 1

    scen_closed = [
        {
            # Spans from Tue 00:00:00 to Wed 23:59:59 999
            "start": "Tue",
            "end": "Wed",
            "expected_start": 1 * NS_IN_DAY,
            "expected_end": 3 * NS_IN_DAY - 1,
        },
        {
            # Spans from Tue 00:00:00 to Wed 23:59:59 999
            "start": "Tuesday",
            "end": "Wednesday",
            "expected_start": 1 * NS_IN_DAY,
            "expected_end": 3 * NS_IN_DAY - 1,
        }
    ]

    scen_open_left = [
        {
            "end": "Tue",
            "expected_end": 2 * NS_IN_DAY - 1 # Tuesday 23:59:59 ...
        }
    ]
    scen_open_right = [
        {
            "start": "Tue",
            "expected_start": 1 * NS_IN_DAY # Tuesday 00:00:00
        }
    ]
    scen_time_point = [
        {
            "start": "Tue",
            "expected_start": 1 * NS_IN_DAY,
            "expected_end": 2 * NS_IN_DAY - 1,
        }
    ]


class TestTimeOfMonth(ConstructTester):

    cls = TimeOfMonth

    max_ns = 31 * NS_IN_DAY - 1

    scen_closed = [
        {
            "start": "2.",
            "end": "3.",
            "expected_start": 1 * NS_IN_DAY,
            "expected_end": 3 * NS_IN_DAY - 1,
        },
        {
            "start": "2nd",
            "end": "4th",
            "expected_start": 1 * NS_IN_DAY,
            "expected_end": 4 * NS_IN_DAY - 1,
        },
    ]

    scen_open_left = [
        {
            "end": "3.",
            "expected_end": 3 * NS_IN_DAY - 1 
        }
    ]
    scen_open_right = [
        {
            "start": "2.",
            "expected_start": 1 * NS_IN_DAY 
        }
    ]
    scen_time_point = [
        {
            "start": "2.",
            "expected_start": 1 * NS_IN_DAY,
            "expected_end": 2 * NS_IN_DAY - 1,
        }
    ]




import pytest
from rocketry.pybox import query

@pytest.mark.parametrize('qry_dict,expected',
    [
        pytest.param(
            {
                'mykey$min': 10,
                'mykey$max': 20,
            },
            {'mykey': (10, 20)},
            id='min-max'
        ),
        pytest.param({'mykey': 10}, {'mykey': 10}, id='equal'),
        pytest.param({'mykey$min': 10}, {'mykey': (10, None)}, id='min'),
        pytest.param({'mykey$max': 10}, {'mykey': (None, 10)}, id='max'),
        pytest.param({'mykey1$min': 10, 'mykey1$max': 20, 'mykey2': 30}, {'mykey1': (10, 20), 'mykey2': 30}, id='combined'),
        pytest.param({}, {}, id='none'),
    ]
)
def test_to_pynative(qry_dict, expected):
    qry = query.from_dict_string(qry_dict)
    kwargs = qry.to_pykwargs()
    assert kwargs == expected



from ..utils.string_parser import InstructionParser
from rocketry.conditions import Any, All, Not
from .condition_item import parse_condition_item

def _flatten(*args, types, with_attr):
    comps = []
    for arg in args:
        if isinstance(arg, types):
            comps += getattr(arg, with_attr)
        else:
            comps.append(arg)
    return comps
def _parse_any(*args):
    comps = _flatten(*args, types=Any, with_attr="subconditions")
    return Any(*comps)

def _parse_all(*args):
    comps = _flatten(*args, types=All, with_attr="subconditions")
    return All(*comps)

def _parse_not(arg):
    if hasattr(arg, "__invert__"):
        return ~arg
    elif isinstance(arg, Not):
        return ~arg
    else:
        return Not(arg)

parse_condition_string = InstructionParser(
    parse_condition_item,
    operators=[
        {
            "symbol": "~",
            "func": _parse_not,
            "side": "right",
        },
        {
            "symbol": "&",
            "func": _parse_all,
            "side": "both",
        },
        {
            "symbol": "|",
            "func": _parse_any,
            "side": "both",
        },
    ]
)



from rocketry.time.interval import (
    TimeOfHour,
    TimeOfDay, 
    TimeOfWeek,
    TimeOfMonth,
    TimeOfYear,
)
from rocketry.time import All, Any

def test_equal():
    assert not (TimeOfHour("10:00") == TimeOfHour("11:00"))
    assert TimeOfHour("10:00") == TimeOfHour("10:00")
    assert (TimeOfHour("10:00", "12:00") & TimeOfHour("11:00", "13:00")) == (TimeOfHour("10:00", "12:00") & TimeOfHour("11:00", "13:00"))
    assert (TimeOfHour("10:00", "12:00") | TimeOfHour("11:00", "13:00")) == (TimeOfHour("10:00", "12:00") | TimeOfHour("11:00", "13:00"))

def test_and():
    time = TimeOfHour("10:00", "14:00") & TimeOfHour("09:00", "12:00")
    assert time == All(TimeOfHour("10:00", "14:00"), TimeOfHour("09:00", "12:00"))

def test_any():
    time = TimeOfHour("10:00", "14:00") | TimeOfHour("09:00", "12:00")
    assert time == Any(TimeOfHour("10:00", "14:00"), TimeOfHour("09:00", "12:00"))



#from rocketry.core.time.base import CLS
from ._time import parse_time_string

from .utils import ParserPicker

def _parse_time_string(s:str, **kwargs):
    time = parse_time_string(s, **kwargs)
    time._str = s
    return time

parse_time = ParserPicker(
    {
        str: _parse_time_string,
    }
) 



import logging
import warnings
import datetime
from typing import TYPE_CHECKING, Iterable, List, Dict, Union

from dateutil.parser import parse as _parse_datetime
from redbird import BaseRepo

from rocketry.core.utils import is_main_subprocess
from rocketry.pybox import query

if TYPE_CHECKING:
    from rocketry.core import Task

class TaskAdapter(logging.LoggerAdapter):
    """Logging adapter for tasks.

    The adapter includes the name of the given 
    task to the log records and allows reading
    the log records if a handler with reading
    capability is found.

    Parameters
    ----------
    logger : logging.Logger
        Logger the TaskAdapter is for.
    task : rocketry.core.Task, str
        Task the adapter is for.
    """
    def __init__(self, logger:logging.Logger, task:Union['Task', str], ignore_warnings=False):
        task_name = task.name if hasattr(task, 'name') else task
        super().__init__(logger, {"task_name": task_name})

        if not ignore_warnings and self.is_readable_unset:
            warnings.warn(f"Logger '{logger.name}' for task '{self.task_name}' does not have ability to be read. Past history of the task cannot be utilized.")

    def process(self, msg, kwargs):
        ""
        kwargs["extra"] = kwargs.get("extra", {})
        kwargs["extra"].update(self.extra)
        return msg, kwargs

    def filter_by(self, *args, **kwargs):
        "Filter by the repo"
        task_name = self.extra["task_name"]
        if task_name is not None:
            kwargs["task_name"] = task_name
        repo = self._get_repo()
        return repo.filter_by(*args, **kwargs)

    def get_records(self, *args, **kwargs) -> Iterable[Dict]:
        r"""Get the log records of the task from the 
        handlers of the logger.
        
        One of the handlers in the logger must 
        have one of the methods:

        - read()
        - query(qry)

        Parameters
        ----------
        qry : list of tuples, dict, Expression
            Query expression to filter records
        **kwargs : dict
            Keyword arguments turned to a query (if qry is None)

        """
        return self.filter_by(*args, **kwargs).all()

    def _get_repo(self) -> BaseRepo:
        "Get repository where the log records are stored"
        handlers = self.logger.handlers
        for handler in handlers:
            repo = getattr(handler, 'repo', None)
            if repo is not None:
                return repo
        else:
            raise AttributeError(f"Logger '{self.logger.name}' has no handlers with repository. Cannot be read.")

    def get_latest(self, action:str=None) -> dict:
        """Get latest log record. Note that this
        is in the same order as in which the 
        handler(s) return the log records.
        
        Parameters
        ----------
        action : str
            Filtering with latest action of this
            type.
        """
        kwargs = {'action': action} if action is not None else {}
        return self.filter_by(**kwargs).last()

# For some reason the logging.Adapter is missing some
# methods that are on logging.Logger
    def handle(self, *args, **kwargs):
        "See `Logger.handle <https://docs.python.org/3/library/logging.html#logging.Logger.handle>`_"
        return self.logger.handle(*args, **kwargs)

    def addHandler(self, *args, **kwargs):
        "See `Logger.addHandler <https://docs.python.org/3/library/logging.html#logging.Logger.addHandler>`_"
        return self.logger.addHandler(*args, **kwargs)

    def __eq__(self, o: object) -> bool:
        is_same_type = type(self) == type(o)
        has_same_logger = self.logger == o.logger
        has_same_name = self.name == o.name
        return is_same_type and has_same_logger and has_same_name

    @property
    def task_name(self):
        return self.extra['task_name']

    @property
    def is_readable(self):
        "bool: Whether the logger is also readable"
        handlers = self.logger.handlers
        for handler in handlers:
            if hasattr(handler, 'repo'):
                return True
        else:
            return False

    @property
    def is_readable_unset(self):
        "bool: Whether the logger is for main process"
        is_process_dummy = self.logger.name.endswith("_process")
        return not self.is_readable and not is_process_dummy and is_main_subprocess()




import logging

from rocketry import Rocketry
from rocketry.args.builtin import Return, Task
from rocketry.conditions import TaskStarted
from rocketry.conditions.api import after_success

from rocketry.conds import (
    false, true,
    daily, time_of_hour,
    after_fail, after_success, after_finish
)

def set_logging_defaults():
    task_logger = logging.getLogger("rocketry.task")
    task_logger.handlers = []
    task_logger.setLevel(logging.WARNING)

def test_app_run():
    set_logging_defaults()

    # Creating app
    app = Rocketry(config={'task_execution': 'main'})

    # Creating some tasks
    @app.task(false)
    def do_never():
        ...

    @app.task(daily)
    def do_daily():
        ...
    
    @app.task(after_success(do_daily))
    def do_after():
        ...

    @app.task(daily & (after_fail(do_never) | time_of_hour.before("10:00") | after_success("do_daily")))
    def do_daily_complex():
        ...

    app.session.config.shut_cond = TaskStarted(task='do_after')
    app.run()

    logger = app.session['do_after'].logger
    assert logger.filter_by(action="success").count() == 1

def test_pipe():
    set_logging_defaults()

    # Creating app
    app = Rocketry(config={'task_execution': 'main'})

    # Creating some tasks
    @app.task(true)
    def do_first():
        return 'hello'

    @app.task(after_success(do_first))
    def do_second(arg=Return(do_first)):
        assert arg == 'hello'

    app.session.config.shut_cond = TaskStarted(task=do_second)
    app.run()

    logger = app.session['do_second'].logger
    assert logger.filter_by(action="success").count() == 1

def test_custom_cond():
    set_logging_defaults()

    # Creating app
    app = Rocketry(config={'task_execution': 'main'})

    # Creating some tasks
    @app.cond('is foo')
    def is_foo(task=Task()):
        assert task.name == "do_things"
        return True

    @app.task(true & is_foo)
    def do_things():
        ...

    app.session.config.shut_cond = TaskStarted(task=do_things)
    app.run()

    logger = app.session['do_things'].logger
    assert logger.filter_by(action="success").count() == 1



import pytest

from rocketry.tasks import FuncTask
from rocketry.core.task import Task
from rocketry.exc import TaskInactionException
from rocketry.conditions import AlwaysFalse, AlwaysTrue

from task_helpers import wait_till_task_finish



Task.use_instance_naming = True


def run_successful_func():
    print("Running func")

def run_failing_func():
    print("Running func")
    raise RuntimeError("Task failed")

def run_inaction():
    raise TaskInactionException()

def run_parametrized(integer, string, optional_float=None):
    assert isinstance(integer, int)
    assert isinstance(string, str)
    assert isinstance(optional_float, float)

def run_parametrized_kwargs(**kwargs):
    assert {} != kwargs
    assert isinstance(kwargs["integer"], int)
    assert isinstance(kwargs["string"], str)
    assert isinstance(kwargs["optional_float"], float)

@pytest.mark.parametrize("execution", ["main", "async", "thread", "process"])
@pytest.mark.parametrize(
    "task_func,expected_outcome,exc_cls",
    [
        pytest.param(
            run_successful_func, 
            "success",
            None,
            id="Success"),
        pytest.param(
            run_failing_func, 
            "fail", 
            RuntimeError,
            id="Failure"),
        pytest.param(
            run_inaction, 
            "inaction", 
            None,
            id="Inaction"),
    ],
)
def test_run(task_func, expected_outcome, exc_cls, execution, session):
    task = FuncTask(
        task_func, 
        name="a task",
        execution=execution
    )

    try:
        task()
    except:
        # failing execution="main"
        if expected_outcome != "fail":
            raise

    # Wait for finish
    wait_till_task_finish(task)

    assert task.status == expected_outcome

    records = list(map(lambda e: e.dict(exclude={'created'}), session.get_task_log()))
    assert [
        {"task_name": "a task", "action": "run"},
        {"task_name": "a task", "action": expected_outcome},
    ] == records


async def run_async_successful():
    ...

async def run_async_fail():
    raise RuntimeError("Failed")

async def run_async_inaction():
    raise TaskInactionException("Did nothing")

@pytest.mark.parametrize("execution", ["main", "async", "thread", "process"])
@pytest.mark.parametrize(
    "task_func,expected_outcome",
    [
        pytest.param(run_async_successful, "success", id="Success"),
        pytest.param(run_async_fail, "fail", id="Failure"),
        pytest.param(run_async_inaction, "inaction", id="Inaction"),
    ],
)
def test_run_async(task_func, expected_outcome, execution, session):

    task = FuncTask(
        task_func, 
        name="a task",
        execution=execution
    )

    try:
        task()
    except:
        # failing execution="main"
        if expected_outcome != "fail":
            raise

    # Wait for finish
    wait_till_task_finish(task)

    assert task.status == expected_outcome

    records = list(map(lambda e: e.dict(exclude={'created'}), session.get_task_log()))
    assert [
        {"task_name": "a task", "action": "run"},
        {"task_name": "a task", "action": expected_outcome},
    ] == records

def test_force_run(tmpdir, session):
    
    # Going to tempdir to dump the log files there
    with tmpdir.as_cwd() as old_dir:

        task = FuncTask(
            run_successful_func, 
            name="task",
            start_cond=AlwaysFalse(),
            execution="main"
        )
        task.force_run = True

        assert bool(task)
        assert bool(task)

        task()
        assert not task.force_run


def test_dependency(tmpdir, session):

    # Going to tempdir to dump the log files there
    with tmpdir.as_cwd() as old_dir:

        task_a = FuncTask(
            run_successful_func, 
            name="task_a", 
            start_cond=AlwaysTrue(),
            execution="main"
        )
        task_b = FuncTask(
            run_successful_func, 
            name="task_b", 
            start_cond=AlwaysTrue(),
            execution="main"
        )
        task_dependent = FuncTask(
            run_successful_func,
            name="task_dependent", 
            start_cond="after task 'task_a' & after task 'task_b'",
            execution="main"
        )
        assert not bool(task_dependent)
        task_a()
        assert not bool(task_dependent)
        task_b()
        assert bool(task_dependent)


# Parametrization
def test_parametrization_runtime(tmpdir, session):
    with tmpdir.as_cwd() as old_dir:

        task = FuncTask(
            run_parametrized, 
            name="a task",
            execution="main"
        )

        task(params={"integer": 1, "string": "X", "optional_float": 1.1, "extra_parameter": "Should not be passed"})

        records = list(map(lambda e: e.dict(exclude={'created'}), session.get_task_log()))
        assert [
            {"task_name": "a task", "action": "run"},
            {"task_name": "a task", "action": "success"},
        ] == records

def test_parametrization_local(tmpdir, session):
    with tmpdir.as_cwd() as old_dir:

        task = FuncTask(
            run_parametrized, 
            name="a task",
            parameters={"integer": 1, "string": "X", "optional_float": 1.1},
            execution="main"
        )

        task()

        records = list(map(lambda e: e.dict(exclude={'created'}), session.get_task_log()))
        assert [
            {"task_name": "a task", "action": "run"},
            {"task_name": "a task", "action": "success"},
        ] == records

def test_parametrization_kwargs(tmpdir, session):
    with tmpdir.as_cwd() as old_dir:

        task = FuncTask(
            run_parametrized_kwargs, 
            name="a task",
            parameters={"integer": 1, "string": "X", "optional_float": 1.1},
            execution="main"
        )

        task()

        records = list(map(lambda e: e.dict(exclude={'created'}), session.get_task_log()))
        assert [
            {"task_name": "a task", "action": "run"},
            {"task_name": "a task", "action": "success"},
        ] == records



from rocketry.core import Scheduler
from rocketry.time import TimeDelta
from rocketry.conditions import SchedulerCycles, SchedulerStarted

def test_scheduler_started(tmpdir, session):
    with tmpdir.as_cwd() as old_dir:

        session.config.shut_cond = ~SchedulerStarted(period=TimeDelta("1 second"))
        session.start()

        assert session.scheduler.n_cycles > 1

def test_scheduler_cycles(tmpdir, session):
    with tmpdir.as_cwd() as old_dir:

        session.config.shut_cond = SchedulerCycles() >= 4
        session.start()

        assert session.scheduler.n_cycles == 4



import datetime
import os
from textwrap import dedent

import pytest

from rocketry import Session
from rocketry.parse import parse_condition, parse_time
from rocketry.session import Config
from rocketry.tasks import FuncTask
from rocketry.core import Task, Scheduler, BaseCondition, BaseArgument, Parameters

def assert_default(session:Session):
    for cls in (Task, Scheduler, BaseCondition, BaseArgument, Parameters):
        assert cls.session is session


def test_empty():
    session = Session()
    session.set_as_default()
    assert session.parameters.to_dict() == {}
    assert session.returns.to_dict() == {}
    assert session.tasks == set()

    config = session.config

    assert isinstance(config, Config)
    assert not config.silence_task_prerun
    assert not config.silence_cond_check
    assert config.task_pre_exist == 'raise'
    assert config.timeout == datetime.timedelta(minutes=30)
    #assert config.task_execution == 'main'

    assert session.env is None
    assert_default(session)

def test_timeout_parse():
    session = Session(config={"timeout": 0.1})
    assert session.config.timeout == datetime.timedelta(seconds=0.1)

    session = Session(config={"timeout": "0.1 seconds"})
    assert session.config.timeout == datetime.timedelta(seconds=0.1)

    session = Session(config={"timeout": datetime.timedelta(seconds=0.1)})
    assert session.config.timeout == datetime.timedelta(seconds=0.1)



import logging
from typing import List, Tuple

import pytest

from rocketry.conditions import (
    TaskStarted, 

    TaskFinished, 
    TaskFailed, 
    TaskSucceeded,

    TaskRunning
)
from rocketry.conditions.task import TaskInacted, TaskTerminated
from rocketry.pybox.time.convert import to_datetime
from rocketry.time import (
    TimeDelta, 
    TimeOfDay
)
from rocketry.tasks import FuncTask

from .test_time import setup_task_state


@pytest.mark.parametrize("cls",
    [
        TaskFailed, TaskSucceeded, TaskFinished, TaskRunning, TaskStarted, TaskInacted, TaskTerminated
    ]
)
def test_logs_not_used_false(session, cls, mock_datetime_now):
    session.config.force_status_from_logs = False
    
    task = FuncTask(
        lambda:None, 
        name="the task",
        execution="main"
    )
    logs = [
        ("2021-01-01 12:00:00", state)
        for state in ("success", "fail", "run", "terminate", "inaction")
    ]
    setup_task_state(mock_datetime_now, logs, task=task)
    cond = cls(task=task)
    assert not cond.observe(session=session)

@pytest.mark.parametrize("cls",
    [
        TaskFailed, TaskSucceeded, TaskFinished, TaskRunning, TaskStarted, TaskInacted, TaskTerminated
    ]
)
def test_logs_not_used_true(session, cls, mock_datetime_now):
    session.config.force_status_from_logs = False
    
    task = FuncTask(
        lambda:None, 
        name="the task",
        execution="main"
    )
    for attr in ("last_run", "last_success", "last_fail", "last_inaction", "last_terminate"):
        setattr(task, attr, to_datetime("2000-01-01 12:00:00"))
    if cls == TaskRunning:
        task.status = "run"
    cond = cls(task=task)
    assert cond.observe(session=session)

@pytest.mark.parametrize("cls",
    [
        TaskFailed, TaskSucceeded, TaskFinished, TaskRunning, TaskStarted, TaskInacted, TaskTerminated
    ]
)
def test_logs_not_used_true_inside_period(session, cls, mock_datetime_now):
    session.config.force_status_from_logs = False
    
    task = FuncTask(
        lambda:None, 
        name="the task",
        execution="main"
    )
    for attr in ("last_run", "last_success", "last_fail", "last_inaction", "last_terminate"):
        setattr(task, attr, to_datetime("2000-01-01 12:00:00"))

    cond = cls(task=task)
    cond = cls(task=task, period=TimeOfDay("07:00", "13:00"))
    if cls == TaskRunning:
        task.status = "run"
    mock_datetime_now("2000-01-01 14:00")
    assert cond.observe(session=session)

@pytest.mark.parametrize("cls",
    [
        TaskFailed, TaskSucceeded, TaskFinished, TaskStarted, TaskInacted, TaskTerminated, TaskRunning
    ]
)
def test_logs_not_used_false_outside_period(session, cls, mock_datetime_now):
    session.config.force_status_from_logs = False
    
    task = FuncTask(
        lambda:None, 
        name="the task",
        execution="main"
    )
    for attr in ("last_run", "last_success", "last_fail", "last_inaction", "last_terminate"):
        setattr(task, attr, to_datetime("2000-01-01 05:00:00"))
    cond = cls(task=task, period=TimeOfDay("07:00", "13:00"))
    mock_datetime_now("2000-01-01 14:00")
    if cls == TaskRunning:
        task.status = "run"
    assert not cond.observe(session=session)

@pytest.mark.parametrize("cls",
    [
        TaskFailed, TaskSucceeded, TaskFinished, TaskStarted, TaskInacted, TaskTerminated
    ]
)
def test_logs_not_used_equal_zero(session, cls, mock_datetime_now):
    session.config.force_status_from_logs = False
    
    task = FuncTask(
        lambda:None, 
        name="the task",
        execution="main"
    )
    logs = [
        ("2021-01-01 12:00:00", state)
        for state in ("success", "fail", "run", "terminate", "inaction")
    ]
    setup_task_state(mock_datetime_now, logs, task=task)
    cond = cls(task=task) == 0
    assert cond.observe(session=session)

@pytest.mark.parametrize("cls",
    [
        TaskFailed, TaskSucceeded, TaskFinished, TaskStarted, TaskInacted, TaskTerminated
    ]
)
def test_logs_used(session, cls, mock_datetime_now):
    session.config.force_status_from_logs = False
    
    task = FuncTask(
        lambda:None, 
        name="the task",
        execution="main"
    )
    logs = [
        ("2021-01-01 12:00:00", state)
        for state in ("success", "fail", "run", "terminate", "inaction")
    ]
    setup_task_state(mock_datetime_now, logs, task=task)
    # Only the latest status is stored
    # thus one cannot determine whether the task has run percisely once
    # without looking to logs.
    mock_datetime_now("2021-01-01 14:00")
    if cls is TaskFinished:
        cond = cls(task=task) == 3
    else:
        cond = cls(task=task) == 1 
    assert cond.observe(session=session)




from typing import Tuple, Dict
import inspect

# Copied from rocketry.pybox\meta\func\func.py

def filter_keyword_args(_func, _params:dict=None, **kwargs):
    """Filter only keyword arguments that the 
    function requires."""
    if _params:
        kwargs.update(_params)
    sig = inspect.signature(_func)

    kw_args = [
        val.name
        for name, val in sig.parameters.items()
        if val.kind in (
            inspect.Parameter.POSITIONAL_OR_KEYWORD, # Normal argument
            inspect.Parameter.KEYWORD_ONLY # Keyword argument
        )
    ]
    return {
        key: val for key, val in kwargs.items()
        if key in kw_args
    }




"""
Test whether a time is inside of an interval.
Also tests conditions.IsPeriod that is essentially a 
wrapper for periods.
"""

import pytest
from rocketry.pybox.time.convert import to_datetime

from rocketry.time.interval import (
    TimeOfHour,
    TimeOfDay, 
    TimeOfWeek,
    TimeOfMonth,
    TimeOfYear,
)
from rocketry.conditions import IsPeriod


true_cases = [
    # TimeOfHour
    {
        "cls": TimeOfHour,
        "dt": "2020-01-01 10:30",
        "start": "15:00",
        "end": "45:00",
    },
    {
        "cls": TimeOfHour,
        "dt": "2020-01-01 10:30",
        "start": "1 quarter",
        "end": "3 quarter",
    },
    {
        "cls": TimeOfHour,
        "dt": "2020-01-01 10:05",
        "start": "45:00",
        "end": "15:00",
    },
    {
        "cls": TimeOfHour,
        "dt": "2020-01-01 10:55",
        "start": "45:00",
        "end": "15:00",
    },

    # TimeOfDay
    {
        "cls": TimeOfDay,
        "dt": "2020-01-01 10:00",
        "start": "09:00",
        "end": "12:00",
        "id": "In period"
    },
    {
        "cls": TimeOfDay,
        "dt": "2020-01-01 23:00",
        "start": "22:00",
        "end": "02:00",
        "id": "In period overnight before mid-night"
    },
    {
        "cls": TimeOfDay,
        "dt": "2020-01-01 01:00",
        "start": "22:00",
        "end": "02:00",
        "id": "In period overnight after mid-night"
    },
    {
        "cls": TimeOfDay,
        "dt": "2020-01-01 09:00",
        "start": "10:00",
        "end": "10:00",
        "id": "Full left"
    },
    {
        "cls": TimeOfDay,
        "dt": "2020-01-01 11:00",
        "start": "10:00",
        "end": "10:00",
        "id": "Full right"
    },
    {
        "cls": TimeOfDay,
        "dt": "2020-01-01 10:00",
        "start": "10:00",
        "end": "10:00",
        "id": "Full mid"
    },
    {
        "cls": TimeOfDay,
        "dt": "2020-01-01 10:00",
        "start": None,
        "end": None,
        "id": "Full"
    },
    {
        "cls": TimeOfDay,
        "dt": "2020-01-01 10:30",
        "start": "10:00",
        "end": None,
        "time_point": True,
        "id": "at"
    },

    # TimeOfWeek
    {
        "cls": TimeOfDay,
        "dt": "2024-01-01 01:00", # 2024 starts from monday
        "start": "Mon",
        "end": "Wed",
    },
    {
        "cls": TimeOfDay,
        "dt": "2024-01-03 01:00", # 2024 starts from monday
        "start": "Mon",
        "end": "Wed",
    },
    {
        "cls": TimeOfWeek,
        "dt": "2024-01-02 01:00", # 2024 starts from monday
        "start": "Tue",
        "end": "Tue",
    },
    {
        "cls": TimeOfWeek,
        "dt": "2024-01-07 01:00", # 2024 starts from monday
        "start": "Tue",
        "end": None,
    },
    {
        "cls": TimeOfWeek,
        "dt": "2024-01-07 01:00", # 2024 starts from monday
        "start": None,
        "end": None,
    },

    # TimeOfMonth
    {
        "cls": TimeOfMonth,
        "dt": "2024-01-05 01:00",
        "start": "5th",
        "end": "10th",
    },
    {
        "cls": TimeOfMonth,
        "dt": "2024-01-10 01:00",
        "start": "5th",
        "end": "10th",
    },

    # TimeOfYear
    {
        "cls": TimeOfYear,
        "dt": "2024-01-01 01:00",
        "start": "Jan",
        "end": "Feb",
    },
    {
        "cls": TimeOfYear,
        "dt": "2024-02-29 23:00",
        "start": "Jan",
        "end": "Feb",
    },
]

false_cases = [
    # TimeOfHour
    {
        "cls": TimeOfHour,
        "dt": "2020-01-01 10:14",
        "start": "15:00",
        "end": "45:00",
    },
    {
        "cls": TimeOfHour,
        "dt": "2020-01-01 10:46",
        "start": "15:00",
        "end": "45:00",
    },
    {
        "cls": TimeOfHour,
        "dt": "2020-01-01 10:14",
        "start": "1 quarter",
        "end": "3 quarter",
    },
    {
        "cls": TimeOfHour,
        "dt": "2020-01-01 10:46",
        "start": "1 quarter",
        "end": "3 quarter",
    },

    # TimeOfDay
    {
        "cls": TimeOfDay,
        "dt": "2020-01-01 08:00",
        "start": "09:00",
        "end": "12:00",
        "id": "Before period"
    },
    {
        "cls": TimeOfDay,
        "dt": "2020-01-01 13:00",
        "start": "09:00",
        "end": "12:00",
        "id": "After period"
    },
    {
        "cls": TimeOfDay,
        "dt": "2020-01-01 21:00",
        "start": "22:00",
        "end": "02:00",
        "id": "Before period overnight"
    },
    {
        "cls": TimeOfDay,
        "dt": "2020-01-01 03:00",
        "start": "22:00",
        "end": "02:00",
        "id": "After period overnight"
    },
    {
        "cls": TimeOfDay,
        "dt": "2020-01-01 09:59",
        "start": "10:00",
        "end": None,
        "time_point": True,
        "id": "left of at"
    },
    {
        "cls": TimeOfDay,
        "dt": "2020-01-01 11:00",
        "start": "10:00",
        "end": None,
        "time_point": True,
        "id": "right of at"
    },

    # IsTimeOfWeek
    {
        "cls": TimeOfWeek,
        "dt": "2024-01-01 23:59:59.999", # 2024 starts from monday
        "start": "Tue",
        "end": "Wed",
    },
    {
        "cls": TimeOfWeek,
        "dt": "2024-01-04 00:00:00.001", # 2024 starts from monday
        "start": "Tue",
        "end": "Wed",
    },
    {
        "cls": TimeOfWeek,
        "dt": "2024-01-01 23:59:59.999", # 2024 starts from monday
        "start": "Tue",
        "end": "Tue",
    },
    {
        "cls": TimeOfWeek,
        "dt": "2024-01-03 01:00", # 2024 starts from monday
        "start": "Tue",
        "end": "Tue",
    },
    {
        "cls": TimeOfWeek,
        "dt": "2024-01-01 01:00", # 2024 starts from monday
        "start": "Tue",
        "end": None,
    },
    {
        "cls": TimeOfWeek,
        "dt": "2024-01-06 01:00", # 2024 starts from monday
        "start": None,
        "end": "Fri",
    },

    # TimeOfMonth
    {
        "cls": TimeOfMonth,
        "dt": "2024-01-04 01:00",
        "start": "5th",
        "end": "10th",
    },
    {
        "cls": TimeOfMonth,
        "dt": "2024-01-11 01:00",
        "start": "5th",
        "end": "10th",
    },
    {
        "cls": TimeOfMonth,
        "dt": "2024-01-10 12:30:00.001",
        "start": "5th",
        "end": "10th 12:30:00",
    },
    {
        "cls": TimeOfMonth,
        "dt": "2024-01-28 23:59:59.999",
        "start": "29th",
        "end": "5th",
    },
    {
        "cls": TimeOfMonth,
        "dt": "2024-01-06 00:00:00",
        "start": "29th",
        "end": "5th",
    },

    # TimeOfYear
    {
        "cls": TimeOfYear,
        "dt": "2024-01-31 23:59:59.999",
        "start": "Feb",
        "end": "Jul",
    },
    {
        "cls": TimeOfYear,
        "dt": "2024-08-01 00:00:00",
        "start": "Feb",
        "end": "Jul",
    },
]

def _to_pyparams(cases):
    return pytest.mark.parametrize(
        "dt,start,end,cls,time_point",
        [
            pytest.param(
                case["dt"],
                case["start"],
                case["end"],
                case["cls"],
                case.get('time_point', None),
                id=f"{case['cls'].__name__}({case['start']}, {case['end']}) --> {case['dt']}{' with time_point' if 'time_point' in case else ''}"
            )
            for case in cases
        ]
    )

@_to_pyparams(true_cases)
def test_in(dt, start, end, cls, time_point):
    dt = to_datetime(dt)
    time = cls(start, end, time_point=time_point)

    assert dt in time

@_to_pyparams(false_cases)
def test_not(dt, start, end, cls, time_point):
    dt = to_datetime(dt)
    time = cls(start, end, time_point=time_point)
    
    assert dt not in time

# Test conditions
@_to_pyparams(true_cases)
def test_true(dt, start, end, cls, time_point, mock_datetime_now):
    mock_datetime_now(dt)
    cond = IsPeriod(cls(start, end, time_point=time_point))
    assert bool(cond)

@_to_pyparams(false_cases)
def test_false(dt, start, end, cls, time_point, mock_datetime_now):
    mock_datetime_now(dt)
    cond = IsPeriod(cls(start, end, time_point=time_point))
    assert not bool(cond)



import datetime
import pytest
from rocketry.time import (
    TimeDelta, TimeOfDay
)

@pytest.mark.parametrize(
    "offset,expected",
    [
        pytest.param(
            "10:20:30",
            datetime.timedelta(hours=10, minutes=20, seconds=30),
            id="String"),
    ],
)
def test_construct(offset, expected):
    time = TimeDelta(offset)
    assert expected == time.past

def test_equal():
    assert TimeDelta("2 days") == TimeDelta("2 days")
    assert not (TimeDelta("2 days") == TimeDelta("3 days"))
    assert not (TimeDelta("2 days") == TimeOfDay("10:00", "12:00"))

def test_repr():
    assert str(TimeDelta("2 days")) == 'past 2 days, 0:00:00'
    assert str(TimeDelta(future="2 days")) == 'next 2 days, 0:00:00'
    assert str(TimeDelta(past="1 days", future="2 days")) == 'past 1 day, 0:00:00 to next 2 days, 0:00:00'
    assert repr(TimeDelta("2 days")) == "TimeDelta(past=datetime.timedelta(days=2), future=datetime.timedelta(0))"